"""
Plot allele usage
"""
import sys
import logging
import pandas as pd
from sqt import SequenceReader
from .table import read_table

logger = logging.getLogger(__name__)


def add_arguments(parser):
	arg = parser.add_argument
	arg('--d-evalue', type=float, default=1E-4,
		help='Maximal allowed E-value for D gene match. Default: %(default)s')
	arg('--d-coverage', '--D-coverage', type=float, default=65,
		help='Minimum D coverage (in percent). Default: %(default)s%%)')
	arg('--database', metavar='FASTA',
		help='Restrict plotting to the sequences named in the FASTA file. '
		     'Only the sequence names are used!')
	arg('--order', metavar='FASTA',
		help='Sort genes according to the order of the records in the FASTA file.')
	arg('--x', choices=('V', 'D', 'J'), default='V',
		help='Type of gene on x axis. Default: %(default)s')
	arg('--gene', choices=('V', 'D', 'J'), default='J',
		help='Type of gene on y axis. Default: %(default)s')
	arg('alleles', help='List of alleles to plot on y axis, separated by comma')
	arg('table', help='Table with parsed and filtered IgBLAST results')
	arg('plot', help='Path to output PDF or PNG')


def main(args):
	usecols = ['V_gene', 'D_gene', 'J_gene', 'V_errors', 'D_errors', 'J_errors', 'D_covered',
		'D_evalue']

	# Support reading a table without D_errors
	try:
		table = read_table(args.table, usecols=usecols)
	except ValueError:
		usecols.remove('D_errors')
		table = read_table(args.table, usecols=usecols)
	logger.info('Table with %s rows read', len(table))
	if args.x == 'V' or args.gene == 'V':
		table = table[table.V_errors == 0]
		logger.info('%s rows remain after requiring V errors = 0', len(table))
	if args.gene == 'J' or args.x == 'J':
		table = table[table.J_errors == 0]
		logger.info('%s rows remain after requiring J errors = 0', len(table))
	elif args.gene == 'D' or args.x == 'D':
		table = table[table.D_evalue <= args.d_evalue]
		logger.info('%s rows remain after requiring D E-value <= %s', len(table), args.d_evalue)
		table = table[table.D_covered >= args.d_coverage]
		logger.info('%s rows remain after requiring D coverage >= %s', len(table), args.d_coverage)
		if 'D_errors' in table.columns:
			table = table[table.D_errors == 0]
			logger.info('%s rows remain after requiring D errors = 0', len(table))

	gene1 = args.x + '_gene'
	gene2 = args.gene + '_gene'
	expression_counts = table.groupby((gene1, gene2)).size().to_frame().reset_index()
	matrix = pd.DataFrame(
		expression_counts.pivot(index=gene1, columns=gene2, values=0).fillna(0), dtype=int)
	# matrix[v_gene,d_gene] gives co-occurrences of v_gene and d_gene
	print('#\n# Expressed genes with counts\n#')
	# The .sum() is along axis=0, that is, the V gene counts are summed up,
	# resulting in counts for each D/J gene
	for g, count in matrix.sum().iteritems():
		print(g, '{:8}'.format(count))

	alleles = args.alleles.split(',')
	for allele in alleles:
		if allele not in matrix.columns:
			logger.error('Allele %s not expressed in this dataset', allele)
			sys.exit(1)

	matrix = matrix.loc[:, alleles]

	if args.database:
		with SequenceReader(args.database) as f:
			x_names = [record.name for record in f if record.name in matrix.index]
		if not x_names:
			logger.error('None of the sequence names in %r were found in the input table',
				args.database)
			sys.exit(1)
		matrix = matrix.loc[x_names, :]

	if args.order:
		with SequenceReader(args.order) as f:
			names = [r.name.partition('*')[0] for r in f]
			gene_order = {name: index for index, name in enumerate(names)}

		matrix['V_gene_tmp'] = pd.Series(matrix.index, index=matrix.index).apply(
			lambda name: name.partition('*')[0]).map(gene_order)
		matrix.sort_values('V_gene_tmp', inplace=True)
		del matrix['V_gene_tmp']

	print('#\n# Allele-specific expression\n#')
	print(matrix)

	if len(alleles) == 2:
		matrix.loc[:, alleles[1]] *= -1

	# remove all-zero rows
	# allele_expressions = allele_expressions[(allele_expressions > 0.001).any(axis=1)]
	ax = matrix.plot(kind='bar', stacked=True, figsize=(12, 6))
	ax.legend(title=None)
	ax.set_title('Allele-specific expression counts')
	ax.set_xlabel(args.x + ' gene')
	ax.set_ylabel('Count')
	ax.figure.set_tight_layout(True)
	# ax.legend(bbox_to_anchor=(1.15, 0.5))
	ax.figure.savefig(args.plot)
	logger.info('Plotted %r', args.plot)



import sys
import os
import subprocess
from contextlib import contextmanager
from io import StringIO


@contextmanager
def capture_stdout():
	sio = StringIO()
	old_stdout = sys.stdout
	sys.stdout = sio
	yield sio
	sys.stdout = old_stdout


def datapath(path):
	return os.path.join(os.path.dirname(__file__), 'data', path)


def resultpath(path):
	return os.path.join(os.path.dirname(__file__), 'results', path)


def files_equal(path1, path2):
	return subprocess.call(['diff', '-u', path1, path2]) == 0




"""
This trie can be used to store a set of strings and to retrieve sequences similar to a query
sequence. 
"""


class Trie:
	"""
	A tree-like datastructure for storing strings. It supports queries similar to a set() of
	strings. This particular implementation also supports searching for strings by Hamming distance.
	"""
	# Children
	A = None
	C = None
	G = None
	T = None
	# The name attribute is set to the string that is spelled from the root of the tree to this
	# node - but only for leaf nodes.
	name = None

	def __init__(self, iterable=None):
		if iterable is not None:
			for it in iterable:
				self.add(it)

	def add(self, s: str):
		"""Add a string to this trie. If it already exists, the trie remains unchanged."""
		self._insert(s, leaf_name=s)

	def _insert(self, s: str, leaf_name: str):
		"""Recursive insert, called by add()"""
		if len(s) == 0:
			# This needs to become a leaf node
			assert self.name is None or self.name == leaf_name
			self.name = leaf_name
		else:
			subtrie = getattr(self, s[0])
			if subtrie is None:
				subtrie = Trie()
				setattr(self, s[0], subtrie)
			subtrie._insert(s[1:], leaf_name)

	def __repr__(self):
		parts = []
		for c in list('ACGT'):
			if getattr(self, c) is not None:
				if getattr(self, c) is None:
					parts.append(c)
				else:
					parts.append(c + ':' + repr(getattr(self, c)))
		if self.name is not None:
			parts.append('Leaf({})'.format(self.name))
		if parts:
			return '{' + ', '.join(parts) + '}'
		else:
			return '*'

	def find_node(self, s: str):
		"""
		Search for s. If s is a prefix of any string in the trie, then return the node
		corresponding to that prefix. The node can be an internal or a leaf node.
		A leaf node is a node in which name attribute is not None. If the returned node
		is a leaf node, then s is in the trie. If it is an internal node (that is,
		the name attribute is None), then there exists a string in the trie that has
		s as a proper prefix.
		
		If no such string (with s as prefix) exists, then None is returned.
		"""
		if len(s) == 0:
			return self
		subtrie = getattr(self, s[0])
		if subtrie is None:
			return None
		else:
			return subtrie.find_node(s[1:])

	def __contains__(self, s: str):
		"""Return whether s is in the trie"""
		node = self.find_node(s)
		return node is not None and node.name == s

	def count_nodes(self, internal=True):
		"""Return number of nodes in this trie. internal nodes are counted if "internal" is True"""
		n = 1 if internal or self.name is not None else 0
		for c in list('ACGT'):
			subtrie = getattr(self, c)
			if subtrie is not None:
				n += len(subtrie)
		return n

	def __len__(self):
		"""Return the number of unique strings in this trie"""
		return self.count_nodes(internal=False)

	def has_similar(self, s, mismatches):
		"""
		Return whether a string exists in this trie that has a Hamming distance
		of at most 'mismatches' to s.
		"""
		if len(s) == 0:
			return self.name is not None
		# As a runtime heuristic, descend into the the mismatches==0 subtrie first
		# if possible.
		subtrie = getattr(self, s[0])
		if subtrie is not None and subtrie.has_similar(s[1:], mismatches):
			return True
		if mismatches == 0:
			return False
		# The above did not work - try all three possible substitutions
		# and recursively check subtries
		for c in list('ACGT'):
			if c == s[0]:
				continue
			subtrie = getattr(self, c)
			if subtrie is None:
				continue
			if subtrie.has_similar(s[1:], mismatches - 1):
				return True
		return False

	def find_all_similar(self, s, mismatches):
		"""
		Yield all strings in this trie that have a Hamming distance of at most
		'mismatches' to s.
		"""
		# This routine is similar to has_similar, but since we are interested
		# in all similar strings, the optimization of going into the matching
		# subtrie first does not help.
		if len(s) == 0:
			if self.name is not None:
				yield self.name
			return

		# The code below is an optimized version of the following:
		#
		# for c in list('ACGT'):
		#   if c != s[0] and mismatches == 0:
		#     continue
		#   subtrie = getattr(self, c)
		#   if subtrie is not None:
		#     yield from subtrie.find_all_similar(s[1:], mismatches - int(c != s[0]))

		if mismatches > 0:
			s0 = s[0]
			s1 = s[1:]
			subtrie = self.A
			if subtrie is not None:
				yield from subtrie.find_all_similar(s1, mismatches - (int(s0 != 'A')))
			subtrie = self.C
			if subtrie is not None:
				yield from subtrie.find_all_similar(s1, mismatches - (int(s0 != 'C')))
			subtrie = self.G
			if subtrie is not None:
				yield from subtrie.find_all_similar(s1, mismatches - (int(s0 != 'G')))
			subtrie = self.T
			if subtrie is not None:
				yield from subtrie.find_all_similar(s1, mismatches - (int(s0 != 'T')))
		else:
			subtrie = self
			for c in s:
				subtrie = getattr(subtrie, c)
				if subtrie is None:
					break
			else:
				if subtrie.name is not None:
					yield subtrie.name



"""
Species-specific code, such as lists of motifs and regular expressions.

Some refactoring is needed to make this module actually usable for many
species. Right now, it works for - at least - human, rhesus monkey and mouse.
"""
import re
from .utils import nt_to_aa


# Regular expressions for CDR3 detection
#
# The idea comes from D’Angelo et al.: The antibody mining toolbox.
# http://dx.doi.org/10.4161/mabs.27105
# The heavy-chain regex was taken directly from there, but the difference
# is that we express everything in terms of amino acids, not nucleotides.
# This simplifies the expressions and makes them more readable.
#
_CDR3_REGEX = {
	# Heavy chain
	'VH': re.compile("""
		[FY] [FHVWY] C
		(?P<cdr3>
			[ADEGIKMNRSTV] .{3,31}
		)
		W[GAV]
		""", re.VERBOSE),

	# Light chain, kappa
	'VK': re.compile("""
		[FSVY] [CFHNVY] [CDFGLSW]
		(?P<cdr3>
			.{4,15}
		)
		[FLV][GRV]
		""", re.VERBOSE),

	# Light chain, lambda
	'VL': re.compile("""
		# the negative lookahead assertion ensures that the rightmost start is found
		[CDY](?![CDY][CFHSY][CFGW])[CFHSY][CFGW]
		(?P<cdr3>
			.{4,15}
		)
		[FS]G
		""", re.VERBOSE)
}

_CDR3_VH_ALTERNATIVE_REGEX = re.compile("""
		C
		(?P<cdr3> . [RK] .{3,30})
		[WF]G.G
""", re.VERBOSE)


def find_cdr3(sequence, chain):
	"""
	Find the CDR3 in the given sequence, assuming it comes from the given chain ('VH', 'VK', 'VL').
	If the chain is not one of 'VH', 'VK', 'VL', return None.

	Return a tuple (start, stop) if found, None otherwise.
	"""
	try:
		regex = _CDR3_REGEX[chain]
	except KeyError:
		return None
	matches = []
	for offset in 0, 1, 2:
		aa = nt_to_aa(sequence[offset:])
		match = regex.search(aa)
		if not match and chain == 'VH':
			match = _CDR3_VH_ALTERNATIVE_REGEX.search(aa)
		if match:
			start, stop = match.span('cdr3')
			matches.append((start * 3 + offset, stop * 3 + offset))
	return min(matches, default=None)


# The following code is used for detecting CDR3 start sites within V
# reference sequences and CDR3 end sites within J reference sequences.


# Matches the start of the CDR3 within the end of a VH sequence
_CDR3_START_VH_REGEX = re.compile("""
	[FY] [FHVWY] C
	(?P<cdr3_start>
		[ADEGIKMNRSTV*] | $
	)
	""", re.VERBOSE)


_CDR3_START_VH_ALTERNATIVE_REGEX = re.compile("""
	C
	(?P<cdr3_start> . [RK])
	""", re.VERBOSE)


_CDR3_START_REGEXES = {
	'kappa': re.compile('[FSVY][CFHNVY][CDFGLSW]'),
	'lambda': re.compile('[CDY](?![CDY][CFHSY][CFGW])[CFHSY][CFGW]'),
	'gamma': re.compile('[YFH]C'),  # TODO test whether this also works for alpha and beta
	'delta': re.compile('[YFH]C'),
}


def _cdr3_start_heavy(aa):
	head, tail = aa[:-15], aa[-15:]
	match = _CDR3_START_VH_REGEX.search(tail)
	if not match:
		match = _CDR3_START_VH_ALTERNATIVE_REGEX.search(tail)
	if not match:
		return None
	return len(head) + match.start('cdr3_start')


def cdr3_start(nt, chain):
	"""
	Find CDR3 start location within a V gene (Ig or TCR)

	nt -- nucleotide sequence of the gene
	chain -- one of the following strings:
	  - 'heavy', 'lambda', 'kappa' for Ig genes
	  - 'alpha', 'beta', 'gamma', 'delta' for TCR genes
	"""
	aa = nt_to_aa(nt)
	if chain == 'heavy':
		start = _cdr3_start_heavy(aa)
		if start is None:
			return None
		return 3 * start
	if chain in ('kappa', 'lambda', 'gamma', 'delta'):
		head, tail = aa[:-15], aa[-15:]
		match = _CDR3_START_REGEXES[chain].search(tail)
		if match:
			return 3 * (len(head) + match.end())
		else:
			return None
	elif chain in ('alpha', 'beta'):
		head, tail = aa[:-8], aa[-8:]
		pos = tail.find('C')
		if pos == -1:
			return None
		else:
			return 3 * (len(head) + pos + 1)


# Matches after the end of the CDR3 within a J sequence
_CDR3_END_REGEXES = {
	'heavy': re.compile('W[GAV]'),
	'kappa': re.compile('FG'),
	'lambda': re.compile('FG'),
	'alpha': re.compile('FG'),
	'beta': re.compile('FG'),
	'gamma': re.compile('FG'),
	'delta': re.compile('FG'),
}


def cdr3_end(nt, chain):
	"""
	Find the position of the CDR3 end within a J sequence

	nt -- nucleotide sequence of the J gene
	chain -- one of the following strings:
	  - 'heavy', 'lambda', 'kappa' for Ig genes
	  - 'alpha', 'beta', 'gamma', 'delta' for TCR genes
	"""
	regex = _CDR3_END_REGEXES[chain]
	for frame in 0, 1, 2:
		aa = nt_to_aa(nt[frame:])
		match = regex.search(aa)
		if match:
			return match.start() * 3 + frame
	return None


# When searching for the CDR3, start this many bases to the left of the end of
# the V match.
CDR3_SEARCH_START = 30



import os
import sys
from tempfile import TemporaryDirectory
import pytest

from igdiscover.__main__ import main
from .utils import datapath, resultpath, files_equal


def run(args, expected):
	"""
	Run IgDiscover, redirecting stdout to a temporary file.
	Then compare the output with the contents of an expected file.
	"""
	with TemporaryDirectory() as td:
		outpath = os.path.join(td, 'output')
		print('Running:', ' '.join(args))
		with open(outpath, 'w') as f:
			old_stdout = sys.stdout
			sys.stdout = f
			main(args)
			sys.stdout = old_stdout
		assert files_equal(expected, outpath)


def test_main():
	with pytest.raises(SystemExit) as exc:
		main(['--version'])
	assert exc.value.code == 0


def test_group_by_barcode_only():
	args = ['group', '-b', '4', datapath('ungrouped.fasta')]
	run(args,  resultpath('grouped-by-barcode-only.fasta'))


def test_group_by_pseudo_cdr3():
	args = ['group', '-b', '4', '--pseudo-cdr3=-5:-2', '--trim-g', datapath('ungrouped.fasta')]
	run(args,  resultpath('grouped.fasta'))


def test_group_by_pseudo_cdr3_barcode_at_end():
	args = ['group', '-b', '-4', '--pseudo-cdr3=1:3', datapath('ungrouped.fasta')]
	run(args, resultpath('grouped2.fasta'))



"""
Filter table with parsed IgBLAST results

Discard the following rows in the table:
- no J assigned
- stop codon found
- V gene coverage less than 90%
- J gene coverage less than 60%
- V gene E-value greater than 1E-3

The filtered table is printed to standard output.
"""
import logging
import pandas as pd

from .table import fix_columns

logger = logging.getLogger(__name__)


def add_arguments(parser):
	arg = parser.add_argument
	arg('--v-coverage', type=float, default=90, metavar='PERCENT',
		help='Require that the sequence covers at least PERCENT of the V gene. '
		'Default: %(default)s')
	arg('--v-evalue', type=float, default=1E-3, metavar='EVALUE',
		help='Require that the E-value for the V gene match is at most EVALUE. '
		'Default: %(default)s')
	arg('--j-coverage', type=float, default=60, metavar='PERCENT',
		help='Require that the sequence covers at least PERCENT of the J gene. '
		'Default: %(default)s')
	arg('table', help='Table with filtered IgBLAST results.')


class FilteringStatistics:
	__slots__ = ('n', 'vjassigned', 'stop', 'v_evalue', 'v_coverage', 'j_coverage')

	def __init__(self):
		self.n = 0
		self.vjassigned = 0
		self.stop = 0
		self.v_evalue = 0
		self.v_coverage = 0
		self.j_coverage = 0

	def __iadd__(self, other):
		for att in self.__slots__:
			v = getattr(self, att)
			setattr(self, att, v + getattr(other, att))
		return self


def filtered_table(table,
		v_gene_coverage,  # at least
		j_gene_coverage,  # at least
		v_gene_evalue,  # at most
	):
	"""
	Discard the following rows in the table:
	- no J assigned
	- stop codon found
	- V gene coverage less than v_gene_coverage
	- J gene coverage less than j_gene_coverage
	- V gene E-value greater than v_gene_evalue

	Return the filtered table.
	"""
	stats = FilteringStatistics()
	stats.n = len(table)
	# Both V and J must be assigned
	# (Note V_gene and J_gene columns use empty strings instead of NA)
	filtered = table[(table['V_gene'] != '') & (table['J_gene'] != '')][:]
	stats.vjassigned = len(filtered)
	filtered['V_gene'] = pd.Categorical(filtered['V_gene'])

	# Filter out sequences that have a stop codon
	filtered = filtered[filtered.stop == 'no']
	stats.stop = len(filtered)

	# Filter out sequences with a too low V gene hit E-value
	filtered = filtered[filtered.V_evalue <= v_gene_evalue]
	stats.v_evalue = len(filtered)

	# Filter out sequences with too low V gene coverage
	filtered = filtered[filtered.V_covered >= v_gene_coverage]
	stats.v_coverage = len(filtered)

	# Filter out sequences with too low J gene coverage
	filtered = filtered[filtered.J_covered >= j_gene_coverage]
	stats.j_coverage = len(filtered)

	return filtered, stats


def main(args):
	n = 0
	first = True
	written = 0
	stats = FilteringStatistics()
	for chunk in pd.read_csv(args.table, chunksize=10000, float_precision='high', sep='\t'):
		fix_columns(chunk)
		n += len(chunk)
		filtered, chunk_stats = filtered_table(chunk, v_gene_coverage=args.v_coverage,
			j_gene_coverage=args.j_coverage, v_gene_evalue=args.v_evalue)
		stats += chunk_stats
		print(filtered.to_csv(sep='\t', index=False, header=first), end='')
		first = False
		written += len(filtered)

	logger.info('%s rows in input table', stats.n)
	logger.info('%s rows have both V and J assignment', stats.vjassigned)
	logger.info('%s of those do not have a stop codon', stats.stop)
	logger.info('%s of those have an E-value of at most %s', stats.v_evalue, args.v_evalue)
	logger.info('%s of those cover the V gene by at least %s%%', stats.v_coverage, args.v_coverage)
	logger.info('%s of those cover the J gene by at least %s%%', stats.j_coverage, args.j_coverage)
	logger.info('%d rows written', written)



"""
Dereplicate sequences and remove barcodes

The difference to the 'group' subcommand is that that one
also computes consensus sequences from groups with identical
barcode/CDR3. This one does not.

The barcode can be in the 5' end or the 3' end of the sequence.

Use --trim-g to remove initial runs of G at the 5' end (artifact from RACE protocol).
These are removed after the barcode is removed.
"""

import sys
import logging
from collections import defaultdict
from itertools import islice

from sqt import SequenceReader


logger = logging.getLogger(__name__)


def add_arguments(parser):
	arg = parser.add_argument
	arg('--limit', default=None, type=int, metavar='N',
		help='Limit processing to the first N reads')
	arg('--trim-g', action='store_true', default=False,
		help="Trim 'G' nucleotides at 5' end")
	arg('--minimum-length', '-l', type=int, default=0,
		help='Minimum sequence length')
	arg('--barcode-length', '-b', type=int, default=0,
		help="Length of barcode. Positive for 5' barcode, negative for 3' barcode. "
		     "Default: %(default)s")
	arg('fastx', metavar='FASTA/FASTQ',
		help='FASTA or FASTQ file (can be gzip-compressed) with sequences')


def main(args):
	barcode_length = args.barcode_length

	too_short = 0
	n = 0
	sequences = defaultdict(list)  # maps sequences to a list of Sequence objects containing them
	with SequenceReader(args.fastx) as f:
		for record in islice(f, 0, args.limit):
			n += 1
			if len(record) < args.minimum_length:
				too_short += 1
				continue
			sequences[record.sequence].append(record)

	n_written = 0
	for records in sequences.values():
		# If there are multiple records with the same sequence, pick the first
		record = records[0]

		if barcode_length >= 0:
			barcode = record.sequence[:barcode_length]
			unbarcoded = record[barcode_length:]
		else:
			barcode = record.sequence[barcode_length:]
			unbarcoded = record[:barcode_length]

		if args.trim_g:
			# The RACE protocol leads to a run of non-template Gs in the beginning
			# of the sequence, after the barcode.
			unbarcoded.sequence = unbarcoded.sequence.lstrip('G')
			if unbarcoded.qualities:
				unbarcoded.qualities = unbarcoded.qualities[-len(unbarcoded.sequence):]

		name = record.name.split(maxsplit=1)[0]
		if name.endswith(';'):
			name = name[:-1]

		if barcode_length:
			print('>{};barcode={};size={};\n{}'.format(name, barcode, len(records),
				unbarcoded.sequence))
		else:
			print('>{};size={};\n{}'.format(name, len(records),
				unbarcoded.sequence))

		n_written += 1

	logger.info('%s sequences processed', n)
	logger.info('%s sequences long enough', n - too_short)
	logger.info('%s dereplicated sequences written', n_written)



"""
Determine haplotypes based on co-occurrences of alleles
"""
import sys
import logging
import pandas as pd
from sqt import SequenceReader
from .table import read_table

logger = logging.getLogger(__name__)


# The second-most expressed allele of a gene must be expressed at at least this
# fraction of the highest-expressed allele in order for the gene to be considered
# heterozygous.
HETEROZYGOUS_THRESHOLD = 0.1

#
EXPRESSED_RATIO = 0.1


def add_arguments(parser):
	arg = parser.add_argument
	arg('--d-evalue', type=float, default=1E-4,
		help='Maximal allowed E-value for D gene match. Default: %(default)s')
	arg('--d-coverage', '--D-coverage', type=float, default=65,
		help='Minimum D coverage (in percent). Default: %(default)s%%)')
	arg('--het', metavar='GENETYPE', choices=('V', 'D', 'J'), default='J',
		help='Use heterozygous GENETYPE genes as anchoring point for haplotyping. '
			'Default: %(default)s')
	arg('--gene', metavar='GENETYPE', choices=('V', 'D', 'J'),
		help='Haplotype genes of this GENETYPE. Default is to haplotype all except '
			'the --het gene')
	arg('--order', metavar='FASTA', default=None,
		help='Sort the output according to the order of the records in '
			'this FASTA file.')
	arg('table', help='Table with parsed and filtered IgBLAST results')


def expression_counts(table, gene_type):
	"""
	Return a DataFrame with gene and allele as the row index and columns 'name' and
	'count'. When 'name' is VH1-1*01, gene would be 'VH1-1' and allele
	would be '01'.
	"""
	counts = table.groupby(gene_type + '_gene').size()
	names, _, alleles = zip(*[s.partition('*') for s in counts.index])
	expressions = pd.DataFrame(
		{'gene': names, 'allele': alleles, 'count': counts, 'name': counts.index},
		columns=['gene', 'allele', 'name', 'count']).set_index(['gene', 'allele'])
	return expressions


def filter_alleles(table):
	"""
	Remove alleles that have too low expression relative to the highest-expressed allele
	"""
	max_expression = table['count'].max()
	return table[table['count'] >= HETEROZYGOUS_THRESHOLD * max_expression]


class GeneMissing(Exception):
	pass


def sorted_haplotype(haplotypes, gene_order):
	# import ipdb; ipdb.set_trace()
	d = {name: i for i, name in enumerate(gene_order)}

	def keyfunc(hap):
		name = hap[0] if hap[0] else hap[1]
		gene = name.partition('*')[0]
		try:
			return d[gene]
		except KeyError:
			raise GeneMissing(gene)

	return sorted(haplotypes, key=keyfunc)


class HeterozygousGene:
	def __init__(self, name, alleles):
		"""
		name -- name of this gene, such as VH4-4
		alleles -- list of its alleles, such as ['VH4-4*01', 'VH4-4*02']
		"""
		self.name = name
		self.alleles = alleles


def compute_coexpressions(table, gene_type1, gene_type2):
	coexpressions = table.groupby(
		(gene_type1 + '_gene', gene_type2 + '_gene')).size().to_frame()
	coexpressions.columns = ['count']
	return coexpressions


def cooccurrences(coexpressions, het_alleles, target_groups):
	"""
	het_alleles -- a two-element list of alleles of a heterozygous gene,
	such as ['IGHJ6*02', 'IGHJ6*03'].
	"""
	assert len(het_alleles) == 2

	haplotype = []
	for target_alleles in target_groups:
		is_expressed_list = []
		names = []
		counts = []
		for target_allele, _ in target_alleles.itertuples(index=False):
			ex = []
			for het_allele in het_alleles:
				try:
					e = coexpressions.loc[(het_allele, target_allele), 'count']
				except KeyError:
					e = 0
				ex.append(e)
			ex_total = sum(ex) + 1  # +1 avoids division by zero
			ratios = [x / ex_total for x in ex]
			is_expressed = [ratio >= EXPRESSED_RATIO for ratio in ratios]
			if is_expressed != [False, False]:
				is_expressed_list.append(is_expressed)
				names.append(target_allele)
				counts.append(ex)
		if len(is_expressed_list) == 1:
			is_expressed = is_expressed_list[0]
			if is_expressed == [True, False]:
				haplotype.append((names[0], '', 'deletion', counts[0]))
			elif is_expressed == [False, True]:
				haplotype.append(('', names[0], 'deletion', counts[0]))
			elif is_expressed == [True, True]:
				haplotype.append((names[0], names[0], 'homozygous', counts[0]))
			else:
				assert False
		elif is_expressed_list == [[True, False], [False, True]]:
			haplotype.append((names[0], names[1], 'heterozygous', (counts[0][0], counts[1][1])))
		elif is_expressed_list == [[False, True], [True, False]]:
			haplotype.append((names[1], names[0], 'heterozygous', (counts[0][1], counts[1][0])))
		else:
			for is_expressed, name, count in zip(is_expressed_list, names, counts):
				haplotype.append((
					name if is_expressed[0] else '',
					name if is_expressed[1] else '',
					'',
					count,
				))
	return haplotype


def main(args):
	if args.het == args.gene:
		logger.error('Gene types given for --het and --gene must not be the same')
		sys.exit(1)
	if args.order is not None:
		with SequenceReader(args.order) as sr:
			gene_order = [r.name for r in sr]
	else:
		gene_order = None
	usecols = ['V_gene', 'D_gene', 'J_gene', 'V_errors', 'D_errors', 'J_errors', 'D_covered',
		'D_evalue']
	# Support reading a table without D_errors
	try:
		table = read_table(args.table, usecols=usecols)
	except ValueError:
		usecols.remove('D_errors')
		table = read_table(args.table, usecols=usecols)
	logger.info('Table with %s rows read', len(table))

	table = table[table.V_errors == 0]
	logger.info('%s rows remain after requiring V errors = 0', len(table))
	table = table[table.J_errors == 0]
	logger.info('%s rows remain after requiring J errors = 0', len(table))
	table = table[table.D_evalue <= args.d_evalue]
	logger.info('%s rows remain after requiring D E-value <= %s', len(table), args.d_evalue)
	table = table[table.D_covered >= args.d_coverage]
	logger.info('%s rows remain after requiring D coverage >= %s', len(table), args.d_coverage)
	if 'D_errors' in table.columns:
		table = table[table.D_errors == 0]
		logger.info('%s rows remain after requiring D errors = 0', len(table))

	# Pre-compute expression levels
	expressions = {gt: expression_counts(table, gt) for gt in ('V', 'D', 'J')}

	for gene_type in ('V', 'D', 'J'):
		logger.info('Heterozygous %s genes:', gene_type)
		for _, group in expressions[gene_type].groupby(level='gene'):
			group = filter_alleles(group)
			if len(group) >= 2:
				logger.info('%s with alleles %s -- Counts: %s',
					group.index[0][0],
					', '.join(group['name']),
					', '.join(str(x) for x in group['count'])
				)

	def count_heterozygous(groups):
		return sum(1 for g in groups if len(g) > 1)

	def filtered_group(gene_type):
		"""Group on gene level, filter out low-expressed genes in each group"""
		result = []
		for _, group in expressions[gene_type].groupby(level='gene'):
			result.append(filter_alleles(group))
		return result

	het_gene_type = args.het

	if args.gene:
		target_gene_types = [args.gene]
	else:
		target_gene_types = {'V': ['D', 'J'], 'D': ['V', 'J'], 'J': ['V', 'D']}[het_gene_type]
	filtered_groups = {gene_type: filtered_group(gene_type) for gene_type in ['V', 'D', 'J']}
	# het_groups and target_groups are expressions grouped by gene (one row is an allele)
	het_groups = filtered_groups[het_gene_type]

	coexpressions_dict = {
		gene_type: compute_coexpressions(table, het_gene_type, gene_type)
		for gene_type in ['V', 'D', 'J']}

	for het_alleles in het_groups:
		het_alleles = list(het_alleles['name'])
		if len(het_alleles) == 1:
			continue
		if len(het_alleles) != 2:
			logger.warning('More than two alleles found: %s. Skipping!',
				', '.join(het_alleles))
			continue

		logger.info('Using heterozygous %s gene alleles %s and %s to haplotype genes',
			het_gene_type, het_alleles[0], het_alleles[1])
		haplotype = []
		for target_gene_type in target_gene_types:
			coexpressions = coexpressions_dict[target_gene_type]
			target_groups = filtered_groups[target_gene_type]
			haplotype.extend(cooccurrences(coexpressions, het_alleles, target_groups))

		if gene_order is not None:
			try:
				haplotype = sorted_haplotype(haplotype, gene_order)
			except GeneMissing as e:
				logger.error('Could not sort genes: gene %r not found in %r',
					str(e), args.order)
				sys.exit(1)

		print('# {}:'.format(' vs '.join(het_alleles)))
		print('haplotype1', 'haplotype2', 'type', 'count1', 'count2', sep='\t')
		for h1, h2, typ, count in haplotype:
			print(h1, h2, typ, count[0], count[1], sep='\t')
		print()



# Tests for CDR3 detection
from igdiscover.species import find_cdr3
from igdiscover.utils import nt_to_aa


def split(s):
	for line in s.split('\n'):
		line = line.strip()
		if line:
			yield line.split()


def assert_cdr3_detection(chain, s):
	for amino_acids, sequence in split(s):
		for offset in range(3):
			target = sequence[offset:]
			match = find_cdr3(target, chain)
			assert match is not None
			assert nt_to_aa(target[match[0]:match[1]]) == amino_acids, (chain, amino_acids, offset)


def test_cdr3_detection_heavy():
	heavy = """
	ARRLHSGSYILFDY CAGGTGACCTTGAAGGAGTCTGGTCCTGCGCTGGTGAAACCCACACAGACCCTCACGCTGACCTGCACCTTCTCTGGGTTCTCACTCAGCACTAGTGGTATGGGTGTGGGCTGGATCCGTCAGCCCTCACGGAAGACCCTGGAGTGGCTTGCACACATTTATTGGAATGATGATAAATACTACAGCACATCGCTGAAGAGCAGGCTCACCATCTCCAAGGACACCTCCAAAAACCAGGTGGTTCTAACAATGACCAACATGGACCCTGTGGACACAGCCACATATTACTGTGCACGGAGACTTCATAGTGGGAGCTACATTCTCTTTGACTACTGGGGCCAGGGAGTCCTGGTCACCGTCTCCTCAGGGAGTGCATCCGCCCCAACCCTTTTCCCCCTCGTCTCCTGTGA

	ARIKWLRSPGYGYFDF CAGGTGACCTTGAAGGAGTCTGGTCCTGCGCTGGTGAGACCCACACAGACCCTCACTCTGACCTGCACCTTCTCTGGGTTCTCAATCAGCACCTCTGGAACAGGTGTGGGCTGGATCCGTCAGCCCCCAGGGAAGGCCCTGGAATGGCTTGCAAGCATTTATTGGACTGATGCTAAATACTATAGCACATCGCAGAAGAGCAGGCTCACCATCTCCAAGGACACCTCCAGAAACCAGGTGATTCTAACAATGACCAACATGGAGCCTGTGGACATAGCCACATATTTCTGTGCACGGATAAAGTGGCTGCGGTCCCCAGGCTATGGATACTTCGATTTCTGGGGCCCTGGCACCCCAATCACCATCTCCTCAGGGAGTGCATCCGCCCCAACCCTTTTCCCCCTCGTCTCCTGTGA

	ARHGIAAAGTHNWFDP TCAGCCGACAAGTCCATCAGCACCGCCTACCTGCAGTGGAGCAGCCTGAAGGCCTCGGACACCGCCATGGATTACTGTGCGAGACATGGGATAGCAGCAGCTGGTACCCACAACTGGTTCGACCCCTGGGGCCAGGGAACCCTGGTCACCGTCTCCTCAGGGAGTGCATCCGCCCCAACCCTTTTCCCCCTCGTCTCCTGTGAGAATTCCCCGTCGGCAGGTTGTT
	"""
	assert_cdr3_detection('VH', heavy)


def test_cdr3_detection_kappa():
	kappa = """
QQYDSSPRT TTCAGTGGCAGTGGAGCAGGGACAGATTTCACTCTCACCATCAGCAGTCTGGAACCTGAGGATGTCGCAACTTACTACTGTCAGCAGTATGATAGCAGCCCCCGGACGTTCGGCGCTGGGACCAAGCTGGAAATCAAACGGAGTGTGCAGAAGCCAACTATCTCCCTCTTCCCTCCATCATCTGAGGAGG

QQYSSYPYT GAGCTGGCCTCGGGAGTCCCAGCTCGCTTCAGTGGGAGTGGGTCAGGGACTTCTTTCTCTCTCACAATCAGCAACGTGGAGGCTGAAGATGTTGCAACCTATTACTGTCAGCAGTATAGCAGTTATCCGTACACGTTCGGCGCAGGGACCAAGCTGGAAATCAAACGGAGTGTGCAGAAGCCAACTATCTCCCTCTTCCCTCCATCATCTGAGGAGG

LQYDSSPYT ATTCTCACCATCAGCAGCCTGCAGCCTGAAGACTTTGCAACTTACTACTGTCTACAGTATGATAGTTCCCCGTACACGTTCGGCGCAGGGACCAAGCTGGAAATCAAACGGAGTGTGCAGAAGCCAACTATCTCCCTCTTCCCTCCATCATCTGAGGAGG

FQYYSGRLT ACAGACTTCACTCTCACCATCAGCAGCCTGCAGCCTGAGGACATTGCAGTTTATTACTGTTTCCAGTATTACAGCGGGAGACTCACGTTCGGAGGAGGGACCCGCTTGGAAATCAAACGGAGTGTGCAGAAGCCAACTATCTCCCTCTTCCCTCCATCATCTGAGGAGG
	"""
	assert_cdr3_detection('VK', kappa)


def test_cdr3_detection_lambda():
	lambda_ = """
LTYHGNSGTFV GGATCCAAAAACCCCTCAGCCAATGCAGGAATTTTGCTCATCTCTGAACTCCAGAATGAGGATGAGGCTGACTATTACTGTCTGACATATCATGGTAATAGTGGTACTTTTGTATTCGGTGGAGGAACCAAGCTGACCGTCCTAGGTCAGCCCAAGTCTGCCCCCACAGTCAGCCTGTTCC

QLWDANSTV ATGGCCACACTGACCATCACTGGCGCCCAGGGTGAGGACGAGGCCGACTATTGCTGTCAGTTGTGGGATGCTAACAGTACTGTGTTCGGTGGAGGAACCACGCTGACCGTCCTAGGTCAGCCCAAGTCTGCCCCCACAGTCAGCCTGTTCCCGCCCTCCTC

GVGYSGGYV GATCGCTACTTAACCATCTCCAACATCCAGCCTGAAGACGAGGCTGACTATTTCTGTGGTGTGGGTTATAGCGGTGGTTATGTATTCGGTGGAGGAACCAAGTTGACCGTCCTAGGTCAGCCCAAGTCTGCTCCCACAGTCAGCCTGTTCCCGCCCTCCTC
	"""
	assert_cdr3_detection('VL', lambda_)



from io import StringIO
import pkg_resources
import pytest

from igdiscover.utils import has_stop, validate_fasta, FastaValidationError, find_overlap, merge_overlapping
from igdiscover.config import Config


def test_has_stop():
	assert has_stop('TAA')
	assert has_stop('TAG')
	assert has_stop('TGA')
	assert has_stop('GGGTGA')

	assert not has_stop('GGG')
	assert not has_stop('TAT')
	assert not has_stop('GGGT')
	assert not has_stop('GGGTA')
	assert not has_stop('TATTG')


def assert_dicts_equal(expected, actual):
	assert expected.keys() == actual.keys()
	for k in expected:
		if hasattr(expected[k], 'keys'):
			assert hasattr(actual[k], 'keys')
			assert_dicts_equal(expected[k], actual[k])
		else:
			assert expected[k] == actual[k], '{}: {} vs {}'.format(k, expected[k], actual[k])


def test_config():
	empty_config = Config(file=StringIO('{}'))
	packaged_config = Config(file=pkg_resources.resource_stream('igdiscover', Config.DEFAULT_PATH))
	# force library name to be equal since it is dynamically determined
	empty_config.library_name = packaged_config.library_name = 'nolib'
	e = empty_config.__dict__
	p = packaged_config.__dict__
	assert_dicts_equal(e, p)
	# assert empty_config == packaged_config


def test_validate_empty_record():
	with pytest.raises(FastaValidationError):
		validate_fasta('tests/data/empty-record.fasta')


def test_validate_duplicate_name():
	with pytest.raises(FastaValidationError):
		validate_fasta('tests/data/duplicate-name.fasta')


def test_validate_duplicate_sequence():
	with pytest.raises(FastaValidationError):
		validate_fasta('tests/data/duplicate-sequence.fasta')


def test_find_overlap():
	assert find_overlap('', '') is None
	assert find_overlap('A', '') is None
	assert find_overlap('ABC', 'X') is None
	assert find_overlap('X', 'ABC') is None
	assert find_overlap('A', 'A') == 0
	assert find_overlap('ABCD', 'A') == 0
	assert find_overlap('A', 'ABC') == 0
	assert find_overlap('AB', 'BD') == 1
	assert find_overlap('ABCDE', 'CDE') == 2
	assert find_overlap('ABCDEFGH', 'CDE') == 2
	assert find_overlap('CDE', 'XABCDEFG') == -3
	assert find_overlap('EFGHI', 'ABCDEFG') == -4


def test_merge_overlapping():
	assert merge_overlapping('', '') is None  # TODO
	assert merge_overlapping('ABC', 'DEF') is None
	assert merge_overlapping('HELLOW', 'LOWORLD') == 'HELLOWORLD'
	assert merge_overlapping('LOWORLD', 'HELLOW') == 'HELLOWORLD'
	assert merge_overlapping('HELLOWORLD', 'LOWO') == 'HELLOWORLD'
	assert merge_overlapping('LOWO', 'HELLOWORLD') == 'HELLOWORLD'



"""
Change configuration file

Example:
    igdiscover config --set iterations 1

"""
import os
import logging
import ruamel.yaml

logger = logging.getLogger(__name__)


class ConfigurationError(Exception):
	pass


class Config:
	DEFAULT_PATH = 'igdiscover.yaml'

	def __init__(self, file):
		# Set some defaults.
		self.debug = False
		self.species = None
		self.sequence_type = 'Ig'
		self.merge_program = 'pear'
		self.flash_maximum_overlap = 300
		self.limit = None  # or an integer
		self.multialign_program = 'muscle-fast'
		self.minimum_merged_read_length = 300
		self.mismatch_penalty = None
		self.barcode_length = 0
		self.barcode_consensus = True
		self.iterations = 1
		self.ignore_j = False
		self.d_coverage = 70
		self.subsample = 1000
		self.stranded = False
		self.forward_primers = None
		self.reverse_primers = None
		self.rename = True
		self.race_g = False
		self.seed = 1
		self.exact_copies = None
		self.preprocessing_filter = dict(v_coverage=90, j_coverage=60, v_evalue=1E-3)
		self.pre_germline_filter = dict(
			unique_cdr3s=2,
			unique_js=2,
			whitelist=True,
			cluster_size=0,
			differences=0,
			allow_stop=True,
			cross_mapping_ratio=0.02,
			clonotype_ratio=0.12,
			exact_ratio=0.12,
			unique_d_ratio=0.3,
			unique_d_threshold=10
		)
		self.germline_filter = dict(
			unique_cdr3s=5,
			unique_js=3,
			whitelist=True,
			cluster_size=100,
			differences=0,
			allow_stop=False,
			cross_mapping_ratio=0.02,
			clonotype_ratio=0.12,
			exact_ratio=0.12,
			unique_d_ratio=0.3,
			unique_d_threshold=10
		)
		self.j_discovery = dict(allele_ratio=0.2, cross_mapping_ratio=0.1, propagate=True)
		self.cdr3_location = 'detect'

		self.read_from(file)

	def read_from(self, file):
		content = file.read()
		new_config = self.make_compatible(ruamel.yaml.safe_load(content))
		for key in ('preprocessing_filter', 'pre_germline_filter', 'germline_filter', 'j_discovery'):
			if key in new_config:
				self.__dict__[key].update(new_config[key])
				del new_config[key]
		self.__dict__.update(new_config)

	@staticmethod
	def make_compatible(config):
		"""
		Convert old-style configuration to new style.
		Raise ConfigurationError if configuration is invalid.
		Return updated config dict.
		"""
		if 'barcode_length' in config:
			raise ConfigurationError(
				'Old-style configuration of barcode length via "barcode_length"'
				'is no longer supported.')
		barcode_length_5prime = config.get('barcode_length_5prime', 0)
		barcode_length_3prime = config.get('barcode_length_3prime', 0)
		if barcode_length_5prime > 0 and barcode_length_3prime > 0:
			raise ConfigurationError(
				'barcode_length_5prime and barcode_length_3prime can currently '
				'not both be greater than zero.')
		if barcode_length_5prime > 0:
			config['barcode_length'] = barcode_length_5prime
		elif barcode_length_3prime > 0:
			config['barcode_length'] = -barcode_length_3prime
		config.pop('barcode_length_5prime', None)
		config.pop('barcode_length_3prime', None)

		if 'seed' in config and config['seed'] is False:
			config['seed'] = None

		for key in ('germline_filter', 'pregermline_filter'):
			if key in config and 'allele_ratio' in config[key]:
				config[key]['clonotype_ratio'] = config[key]['allele_ratio']
		return config

	@classmethod
	def from_default_path(cls):
		with open(cls.DEFAULT_PATH) as f:
			return Config(file=f)


def add_arguments(parser):
	arg = parser.add_argument
	arg('--set', nargs=2, default=[], metavar=('KEY', 'VALUE'), action='append',
		help='Set KEY to VALUE. Use KEY.SUBKEY[.SUBSUBKEY...] for nested keys.')
	arg('--file', default=Config.DEFAULT_PATH,
		help='Configuration file to modify. Default: igdiscover.yaml in current directory.')


def main(args):
	if args.set:
		with open(args.file) as f:
			config = ruamel.yaml.load(f, ruamel.yaml.RoundTripLoader)
		for k, v in args.set:
			v = ruamel.yaml.safe_load(v)
			# config[k] = v
			item = config
			# allow nested keys
			keys = k.split('.')
			for i in keys[:-1]:
				item = item[i]
			item[keys[-1]] = v
		tmpfile = args.file + '.tmp'
		with open(tmpfile, 'w') as f:
			print(ruamel.yaml.dump(config, Dumper=ruamel.yaml.RoundTripDumper), end='', file=f)
		os.rename(tmpfile, args.file)
	else:
		with open(args.file) as f:
			config = ruamel.yaml.safe_load(f)
		print(ruamel.yaml.dump(config), end='')



"""
Group assigned sequences by clonotype

Two sequences have the same clonotype if
- their V and J assignments are the same
- the length of their CDR3 is identical
- the difference between their CDR3s (in terms of mismatches)
  is not higher than a given threshold (by default 1)

The output is a table with one row per clonotype, written to
standard output.

Optionally, a full table of all members (sequences belonging to a clonotype)
can be created with one row per input sequence, sorted by
clonotype, plus an empty line between each group of sequences
that have the same clonotype.

The tables are by default sorted by clonotype, but can instead be sorted
by the group size (number of members of a clonotype).
"""
import logging
from itertools import islice
from contextlib import ExitStack
from collections import Counter

import pandas as pd
from xopen import xopen
from sqt.align import hamming_distance

from .table import read_table
from .cluster import hamming_single_linkage
from .utils import slice_arg


CLONOTYPE_COLUMNS = ['name', 'count', 'V_gene', 'D_gene', 'J_gene', 'CDR3_nt', 'CDR3_aa',
	'FR1_SHM', 'CDR1_SHM', 'FR2_SHM', 'CDR2_SHM', 'FR3_SHM',
	'FR1_aa_mut', 'CDR1_aa_mut', 'FR2_aa_mut', 'CDR2_aa_mut', 'FR3_aa_mut',
	'V_errors', 'J_errors', 'V_SHM', 'J_SHM', 'barcode', 'VDJ_nt', 'VDJ_aa']


logger = logging.getLogger(__name__)


def add_arguments(parser):
	arg = parser.add_argument
	arg('--sort', action='store_true', default=False,
		help='Sort by group size (largest first). Default: Sort by V/D/J gene names')
	arg('--limit', metavar='N', type=int, default=None,
		help='Print out only the first N groups')
	arg('--cdr3-core', default=None,
		type=slice_arg, metavar='START:END',
		help='START:END defines the non-junction region of CDR3 '
			'sequences. Use negative numbers for END to count '
			'from the end. Regions before and after are considered to '
			'be junction sequence, and for two CDR3s to be considered '
			'similar, at least one of the junctions must be identical. '
			'Default: no junction region.')
	arg('--mismatches', default=1, type=float,
		help='No. of allowed mismatches between CDR3 sequences. '
			'Can also be a fraction between 0 and 1 (such as 0.15), '
			'interpreted relative to the length of the CDR3 (minus the front non-core). '
			'Default: %(default)s')
	arg('--aa', default=False, action='store_true',
		help='Count CDR3 mismatches on amino-acid level. Default: Compare nucleotides.')
	arg('--members', metavar='FILE',
		help='Write member table to FILE')
	arg('table', help='Table with parsed and filtered IgBLAST results')


def is_similar_with_junction(s, t, mismatches, cdr3_core):
	"""
	Return whether strings s and t have at most the given number of mismatches
	*and* have at least one identical junction.
	"""
	# TODO see issue #81
	if len(s) != len(t):
		return False
	if 0 < mismatches < 1:
		delta = cdr3_core.start if cdr3_core is not None else 0
		distance_ok = hamming_distance(s, t) <= (len(s) - delta) * mismatches
	else:
		distance_ok = hamming_distance(s, t) <= mismatches
	if cdr3_core is None:
		return distance_ok
	return distance_ok and (
			(s[:cdr3_core.start] == t[:cdr3_core.start]) or
			(s[cdr3_core.stop:] == t[cdr3_core.stop:]))


def group_by_cdr3(table, mismatches, cdr3_core, cdr3_column):
	"""
	Cluster the rows of the table by Hamming distance between
	their CDR3 sequences. Yield (index, group) tuples similar 
	to .groupby().
	"""
	# Cluster all unique CDR3s by Hamming distance
	sequences = list(set(table[cdr3_column]))

	def linked(s, t):
		return is_similar_with_junction(s, t, mismatches, cdr3_core)

	clusters = hamming_single_linkage(sequences, mismatches, linked=linked)

	# Create dict that maps CDR3 sequences to a numeric cluster id
	cluster_ids = dict()
	for cluster_id, cdr3s in enumerate(clusters):
		for cdr3 in cdr3s:
			cluster_ids[cdr3] = cluster_id

	# Assign cluster id to each row
	table['cluster_id'] = table[cdr3_column].apply(lambda cdr3: cluster_ids[cdr3])
	for index, group in table.groupby('cluster_id'):
		yield group.drop('cluster_id', axis=1)


def representative(table):
	"""
	Given a table with members of the same clonotype, return a representative
	as a dict.
	"""
	c = Counter()
	for row in table.itertuples():
		c[row.VDJ_nt] += row.count
	most_common_vdj_nt = c.most_common(1)[0][0]
	result = table[table['VDJ_nt'] == most_common_vdj_nt].iloc[0]
	result.at['count'] = table['count'].sum()
	return result


def group_by_clonotype(table, mismatches, sort, cdr3_core, cdr3_column):
	"""
	Yield clonotype groups. Each item is a DataFrame with all the members of the
	clonotype.
	"""
	logger.info('Computing clonotypes ...')
	prev_v = None
	groups = []
	for (v_gene, j_gene, cdr3_length), vj_group in table.groupby(
			('V_gene', 'J_gene', 'CDR3_length')):
		if prev_v != v_gene:
			logger.info('Processing %s', v_gene)
		prev_v = v_gene
		cdr3_groups = group_by_cdr3(vj_group.copy(), mismatches=mismatches, cdr3_core=cdr3_core,
			cdr3_column=cdr3_column)
		if sort:
			# When sorting by group size is requested, we need to buffer
			# results
			groups.extend(cdr3_groups)
		else:
			yield from cdr3_groups

	if sort:
		logger.info('Sorting by group size ...')
		groups.sort(key=len, reverse=True)
		yield from groups


def main(args):
	logger.info('Reading input table ...')
	usecols = CLONOTYPE_COLUMNS
	# TODO backwards compatibility
	if 'FR1_aa_mut' not in pd.read_csv(args.table, nrows=0, sep='\t').columns:
		usecols = [col for col in usecols if not col.endswith('_aa_mut')]

	table = read_table(args.table, usecols=usecols)
	table = table[usecols]
	logger.info('Read table with %s rows', len(table))
	table.insert(5, 'CDR3_length', table['CDR3_nt'].apply(len))
	table = table[table['CDR3_length'] > 0]
	table = table[table['CDR3_aa'].map(lambda s: '*' not in s)]
	logger.info('After discarding rows with unusable CDR3, %s remain', len(table))
	with ExitStack() as stack:
		if args.members:
			members_file = stack.enter_context(xopen(args.members, 'w'))
		else:
			members_file = None

		columns = usecols[:]
		columns.remove('barcode')
		columns.remove('count')
		columns.insert(0, 'count')
		columns.insert(columns.index('CDR3_nt'), 'CDR3_length')
		print(*columns, sep='\t')
		print_header = True
		n = 0
		cdr3_column = 'CDR3_aa' if args.aa else 'CDR3_nt'
		grouped = group_by_clonotype(table, args.mismatches, args.sort, args.cdr3_core, cdr3_column)
		for group in islice(grouped, 0, args.limit):
			if members_file:
				# We get an intentional empty line between groups since
				# to_csv() already includes a line break
				print(group.to_csv(sep='\t', header=print_header, index=False), file=members_file)
				print_header = False
			rep = representative(group)
			print(*[rep[col] for col in columns], sep='\t')
			n += 1
	logger.info('%d clonotypes written', n)



"""
Query a table of assigned sequences by clonotype

Two sequences have the same clonotype if
- their V and J assignments are the same
- the length of their CDR3 is identical
- the difference between their CDR3s (in terms of mismatches)
  is not higher than a given threshold (by default 1)

Clonotypes for the query sequences are determined and sequences
in the input table that have this clonotype are reported.

The table is written to standard output.
"""
import logging
from collections import defaultdict
from contextlib import ExitStack

import pandas as pd
from xopen import xopen

from .table import read_table
from .utils import slice_arg
from .clonotypes import is_similar_with_junction, CLONOTYPE_COLUMNS

logger = logging.getLogger(__name__)


def add_arguments(parser):
	arg = parser.add_argument
	arg('--minimum-count', '-c', metavar='N', default=1, type=int,
		help='Discard all rows with count less than N. Default: %(default)s')
	arg('--cdr3-core', default=None,
		type=slice_arg, metavar='START:END',
		help='START:END defines the non-junction region of CDR3 '
			'sequences. Use negative numbers for END to count '
			'from the end. Regions before and after are considered to '
			'be junction sequence, and for two CDR3s to be considered '
			'similar, at least one of the junctions must be identical. '
			'Default: no junction region.')
	arg('--mismatches', default=1, type=float,
		help='No. of allowed mismatches between CDR3 sequences. '
			'Can also be a fraction between 0 and 1 (such as 0.15), '
			'interpreted relative to the length of the CDR3 (minus the front non-core). '
			'Default: %(default)s')
	arg('--aa', default=False, action='store_true',
		help='Count CDR3 mismatches on amino-acid level. Default: Compare nucleotides.')
	arg('--summary', metavar='FILE',
		help='Write summary table to FILE')
	arg('reftable', help='Reference table with parsed and filtered '
		'IgBLAST results (filtered.tab)')
	arg('querytable', help='Query table with IgBLAST results (assigned.tab or filtered.tab)')


def collect(querytable, reftable, mismatches, cdr3_core_slice, cdr3_column):
	"""
	Find all queries from the querytable in the reftable.

	Yield tuples (query_rows, similar_rows) where the query_rows is a list
	with all the rows that have the same result. similar_rows is a DataFrame
	whose rows are the ones matching the query.
	"""
	# The vjlentype is a "clonotype without CDR3 sequence" (only V, J, CDR3 length)
	# Determine set of vjlentypes to query
	query_vjlentypes = defaultdict(list)
	for row in querytable.itertuples():
		vjlentype = (row.V_gene, row.J_gene, len(row.CDR3_nt))
		query_vjlentypes[vjlentype].append(row)

	groupby = ('V_gene', 'J_gene', 'CDR3_length')
	for vjlentype, vjlen_group in reftable.groupby(groupby):
		# (v_gene, j_gene, cdr3_length) = vjlentype
		if vjlentype not in query_vjlentypes:
			continue

		# Collect results for this vjlentype. The results dict
		# maps row indices (into the vjlen_group) to each query_row,
		# allowing us to group identical results together.
		results = defaultdict(list)
		for query_row in query_vjlentypes.pop(vjlentype):
			cdr3 = getattr(query_row, cdr3_column)
			# Save indices of the rows that are similar to this query
			indices = tuple(index for index, r in enumerate(vjlen_group.itertuples())
				if is_similar_with_junction(cdr3, getattr(r, cdr3_column), mismatches, cdr3_core_slice))
			results[indices].append(query_row)

		# Yield results, grouping queries that lead to the same result
		for indices, query_rows in results.items():
			if not indices:
				for query_row in query_rows:
					yield ([query_row], [])
				continue

			similar_group = vjlen_group.iloc[list(indices), :]
			yield (query_rows, similar_group)

	# Yield result tuples for all the queries that have not been found
	for queries in query_vjlentypes.values():
		for query_row in queries:
			yield ([query_row], [])


def main(args):
	usecols = CLONOTYPE_COLUMNS
	# TODO backwards compatibility
	if ('FR1_aa_mut' not in pd.read_csv(args.querytable, nrows=0, sep='\t').columns or
			'FR1_aa_mut' not in pd.read_csv(args.reftable, nrows=0, sep='\t').columns):
		usecols = [col for col in usecols if not col.endswith('_aa_mut')]
	querytable = read_table(args.querytable, usecols=usecols)
	querytable = querytable[usecols]  # reorder columns
	# Filter empty rows (happens sometimes)
	querytable = querytable[querytable.V_gene != '']
	logger.info('Read query table with %s rows', len(querytable))
	reftable = read_table(args.reftable, usecols=usecols)
	reftable = reftable[usecols]
	logger.info('Read reference table with %s rows', len(reftable))
	if args.minimum_count > 1:
		reftable = reftable[reftable['count'] >= args.minimum_count]
		logger.info('After filtering out rows with count < %s, %s rows remain', args.minimum_count,
			len(reftable))
	for tab in querytable, reftable:
		tab.insert(5, 'CDR3_length', tab['CDR3_nt'].apply(len))

	if len(querytable) > len(reftable):
		logger.warning('The reference table is smaller than the '
			'query table! Did you swap query and reference?')

	cdr3_column = 'CDR3_aa' if args.aa else 'CDR3_nt'
	summary_columns = ['FR1_SHM', 'CDR1_SHM', 'FR2_SHM', 'CDR2_SHM', 'FR3_SHM', 'V_SHM', 'J_SHM',
		'V_aa_mut', 'J_aa_mut']
	summary_columns.extend(col for col in usecols if col.endswith('_aa_mut'))
	with ExitStack() as stack:
		if args.summary:
			summary_file = stack.enter_context(xopen(args.summary, 'w'))
			print('name', 'size', *('avg_' + s for s in summary_columns), sep='\t', file=summary_file)
		else:
			summary_file = None

		# Print header
		print(*reftable.columns, sep='\t')

		# Do the actual work
		for query_rows, result_table in collect(querytable, reftable, args.mismatches,
				args.cdr3_core, cdr3_column):
			assert len(query_rows) >= 1
			if summary_file:
				for query_row in query_rows:
					print(query_row.name, len(result_table), sep='\t', end='', file=summary_file)
					for col in summary_columns:
						mean = result_table[col].mean() if len(result_table) > 0 else 0
						print('\t{:.2f}'.format(mean), end='', file=summary_file)
					print(file=summary_file)

			for query_row in query_rows:
				print('# Query: {}'.format(query_row.name), '', *(query_row[3:]), sep='\t')
			if len(result_table) > 0:
				print(result_table.to_csv(sep='\t', header=False, index=False))
			else:
				print()



"""
Compute union of sequences in multiple FASTA files
"""
import logging
from collections import namedtuple
from sqt import FastaReader

from .utils import Merger

logger = logging.getLogger(__name__)


def add_arguments(parser):
	arg = parser.add_argument
	# arg('--max-differences', type=int, metavar='MAXDIFF', default=0,
	# 	help='Merge sequences if they have at most MAXDIFF differences. '
	# 	' Default: %(default)s')
	arg('fasta', help='FASTA file', nargs='+')


SequenceInfo = namedtuple('SequenceInfo', 'sequence name')


class SequenceMerger(Merger):
	"""
	Merge sequences where one is a prefix of the other into single entries.
	"""
	def __init__(self):
		super().__init__()

	def merged(self, s, t):
		"""
		Merge two sequences if one is the prefix of the other. If they should
		not be merged, None is returned.

		s and t must have attributes sequence and name.
		"""
		s_seq = s.sequence
		t_seq = t.sequence
		# Make both sequences the same length - cheap trick to not penalize
		# end gaps
		s_seq += t_seq[len(s_seq):]
		t_seq += s_seq[len(t_seq):]
		if s_seq == t_seq:
			return s
		return None


def main(args):
	merger = SequenceMerger()
	n_read = 0
	for path in args.fasta:
		n = 0
		for record in FastaReader(path):
			merger.add(SequenceInfo(record.sequence.upper(), record.name))
			n += 1
		n_read += n
		logger.info('Read %s sequences from %s', n, path)
	logger.info('Read %s sequences from %s files', n_read, len(args.fasta))

	for info in merger:
		print('>{}\n{}'.format(info.name, info.sequence))



"""
Parse IgBLAST output and write out a tab-separated table.

IgBLAST must have been run with -outfmt "7 sseqid qstart qseq sstart sseq pident slen"

A few extra things are done in addition to parsing:
- The CDR3 is detected by using a regular expression
- The leader is detected within the sequence before the found V gene (by
  searching for the start codon).
- If the V sequence hit starts at base 2 in the reference, it is extended
  one to the left.
"""
import csv
import logging
from collections import namedtuple

from sqt.dna import reverse_complement
from sqt.align import edit_distance

from .utils import nt_to_aa
from .species import find_cdr3, CDR3_SEARCH_START

logger = logging.getLogger(__name__)


def none_if_na(s):
	"""Return None if s == 'N/A'. Return s otherwise."""
	return None if s == 'N/A' else s


def split_by_section(iterable, section_starts):
	"""
	Parse a stream of lines into chunks of sections. When one of the lines
	starts with a string given in section_starts, a new section is started, and
	a tuple (head, lines) is returned where head is the matching line and lines
	contains a list of the lines following the section header, up to (but
	excluding) the next section header.

	Works a bit like str.split(), but on lines.
	"""
	lines = None
	header = None
	for line in iterable:
		line = line.strip()
		for start in section_starts:
			if line.startswith(start):
				if header is not None:
					yield (header, lines)
				header = line
				lines = []
				break
		else:
			if header is None:
				raise ParseError("Expected a line starting with one of {}".format(', '.join(section_starts)))
			lines.append(line)
	if header is not None:
		yield (header, lines)


AlignmentSummary = namedtuple('AlignmentSummary', 'start stop length matches mismatches gaps percent_identity')
JunctionVDJ = namedtuple('JunctionVDJ', 'v_end vd_junction d_region dj_junction j_start')
JunctionVJ = namedtuple('JunctionVJ', 'v_end vj_junction j_start')


_Hit = namedtuple('_Hit', [
	'subject_id',  # name of database record, such as "VH4.11"
	'query_start',
	'query_alignment',  # aligned part of the query, with '-' for deletions
	'subject_start',
	'subject_alignment',  # aligned reference, with '-' for insertions
	'subject_length',  # total length of reference, depends only on subject_id
	'percent_identity',
	'evalue',
])


class Hit(_Hit):
	# This avoids having a __dict__ attribute, which is necessary for namedtuple
	# subclasses that need _asdict() to work (http://bugs.python.org/issue24931)
	__slots__ = ()

	def covered(self):
		"""
		Return fraction of bases in the original subject sequence that are
		covered by this hit.
		"""
		return len(self.subject_sequence) / self.subject_length

	@property
	def query_end(self):
		return self.query_start + len(self.query_sequence)

	@property
	def subject_end(self):
		return self.subject_start + len(self.subject_sequence)

	@property
	def query_sequence(self):
		return self.query_alignment.replace('-', '')

	@property
	def subject_sequence(self):
		return self.subject_alignment.replace('-', '')

	@property
	def errors(self):
		return sum(a != b for a, b in zip(self.subject_alignment, self.query_alignment))

	def query_position(self, reference_position):
		"""
		Given a position on the reference, return the same position but relative to
		the full query sequence.
		"""
		# Iterate over alignment columns
		ref_pos = self.subject_start
		query_pos = self.query_start
		if ref_pos == reference_position:
			return query_pos
		for ref_c, query_c in zip(self.subject_alignment, self.query_alignment):
			if ref_c != '-':
				ref_pos += 1
			if query_c != '-':
				query_pos += 1
			if ref_pos == reference_position:
				return query_pos
		return None


def parse_header(header):
	"""
	Extract size= and barcode= fields from the FASTA/FASTQ header line

	>>> parse_header("name;size=12;barcode=ACG;")
	('name', 12, 'ACG')
	>>> parse_header("another name;size=200;foo=bar;")
	('another name', 200, None)
	"""
	fields = header.split(';')
	query_name = fields[0]
	size = barcode = None
	for field in fields[1:]:
		if field == '':
			continue
		if '=' in field:
			key, value = field.split('=', maxsplit=1)
			if key == 'size':
				size = int(value)
			elif key == 'barcode':
				barcode = value
	return query_name, size, barcode


class IgBlastRecord:
	def __init__(
		self,
		full_sequence,
		query_name,
		alignments,
		hits,
		v_gene,
		d_gene,
		j_gene,
		chain,
		has_stop,
		in_frame,
		is_productive,
		strand,
		junction
	):
		self.full_sequence = full_sequence
		self.query_name = query_name
		self.alignments = alignments
		self.hits = hits
		self.v_gene = v_gene
		self.d_gene = d_gene
		self.j_gene = j_gene
		self.chain = chain
		self.has_stop = has_stop
		self.in_frame = in_frame
		self.is_productive = is_productive
		self.strand = strand
		self.junction = junction

	def region_sequence(self, region):
		"""
		Return the nucleotide sequence of a named region. Allowed names are:
		CDR1, CDR2, CDR3, FR1, FR2, FR3. For all regions except CDR3, sequences
		are extracted from the full read using begin and end coordinates from
		IgBLAST’s "alignment summary" table.
		"""
		alignment = self.alignments.get(region, None)
		if alignment is None:
			return None
		if alignment.start is None or alignment.stop is None:
			return None
		return self.full_sequence[alignment.start:alignment.stop]

	def __repr__(self):
		return 'IgBlastRecord(query_name={query_name!r}, ' \
			'v_gene={v_gene!r}, d_gene={d_gene!r}, j_gene={j_gene!r}, chain={chain!r}, ...)'.format(
				**vars(self))


class ExtendedIgBlastRecord(IgBlastRecord):
	"""
	This extended record does a few extra things:
	- The CDR3 is detected by using a regular expression
	- The leader is detected within the sequence before the found V gene (by
	searching for the start codon).
	- If the V sequence hit starts at base 2 in the reference, it is extended
	  one to the left.
	"""
	# TODO move computation of cdr3_sequence, vdj_sequence into constructor
	# TODO maybe make all coordinates relative to full sequence

	# Order of columns (use with asdict())
	columns = [
		'count',
		'V_gene',
		'D_gene',
		'J_gene',
		'chain',
		'stop',
		'V_covered',
		'D_covered',
		'J_covered',
		'V_evalue',
		'D_evalue',
		'J_evalue',
		'FR1_SHM',
		'CDR1_SHM',
		'FR2_SHM',
		'CDR2_SHM',
		'FR3_SHM',
		'V_SHM',
		'J_SHM',
		'V_aa_mut',
		'J_aa_mut',
		'FR1_aa_mut',
		'CDR1_aa_mut',
		'FR2_aa_mut',
		'CDR2_aa_mut',
		'FR3_aa_mut',
		'V_errors',
		'D_errors',
		'J_errors',
		'UTR',
		'leader',
		'CDR1_nt',
		'CDR1_aa',
		'CDR2_nt',
		'CDR2_aa',
		'CDR3_nt',
		'CDR3_aa',
		'V_nt',
		'V_aa',
		'V_end',
		'V_CDR3_start',
		'VD_junction',
		'D_region',
		'DJ_junction',
		'J_nt',
		'VDJ_nt',
		'VDJ_aa',
		'name',
		'barcode',
		'genomic_sequence',
	]

	CHAINS = {
		'VH': 'heavy', 'VK': 'kappa', 'VL': 'lambda',
		'VA': 'alpha', 'VB': 'beta',
		'VG': 'gamma', 'VD': 'delta'
	}

	def __init__(self, database, **kwargs):
		super().__init__(**kwargs)
		self.query_name, self.size, self.barcode = parse_header(self.query_name)
		self.genomic_sequence = self.full_sequence
		self._database = database
		if 'V' in self.hits:
			self.hits['V'] = self._fixed_v_hit()
		self.utr, self.leader = self._utr_leader()
		self.alignments['CDR3'] = self._find_cdr3()

	@property
	def vdj_sequence(self):
		if 'V' not in self.hits or 'J' not in self.hits:
			return None
		hit_v = self.hits['V']
		hit_j = self.hits['J']
		vdj_start = hit_v.query_start
		vdj_stop = hit_j.query_start + len(hit_j.query_sequence)
		return self.full_sequence[vdj_start:vdj_stop]

	@property
	def v_cdr3_start(self):
		"""Start of CDR3 within V"""
		if 'V' not in self.hits or self.alignments['CDR3'] is None:
			return 0
		v_start = self.hits['V'].query_start
		cdr3_start = self.alignments['CDR3'].start
		return cdr3_start - v_start

	def _utr_leader(self):
		"""
		Split the sequence before the V gene match into UTR and leader by
		searching for the start codon.
		"""
		if 'V' not in self.hits:
			return None, None
		before_v = self.full_sequence[:self.hits['V'].query_start]

		# Search for the start codon
		for offset in (0, 1, 2):
			for i in range(66, 42, -3):
				if before_v[-i + offset : -i + 3 + offset] == 'ATG':
					return before_v[:-i + offset], before_v[-i + offset:]
		return None, None

	# TODO this is unused
	def _fixed_cdr3_alignment_by_regex(self):
		"""
		Return a repaired AlignmentSummary object for the CDR3 region which
		does not use IgBLAST’s coordinates. IgBLAST does not determine the end
		of the CDR3 correctly, at least when a custom database is used,

		Return (start, end) of CDR3 relative to query. The CDR3 is detected
		using a regular expression. Return None if no CDR3 detected.
		"""
		if 'V' not in self.hits or 'J' not in self.hits:
			return None
		# Search in a window around the V(D)J junction for the CDR3
		if 'CDR3' in self.alignments:
			window_start = self.alignments['CDR3'].start - CDR3_SEARCH_START
		else:
			window_start = max(0, self.hits['V'].query_end - CDR3_SEARCH_START)
		window_end = self.hits['J'].query_end
		window = self.full_sequence[window_start:window_end]
		match = find_cdr3(window, self.chain)
		if not match:
			return None
		start = match[0] + window_start
		end = match[1] + window_start
		assert start < end
		return AlignmentSummary(start=start, stop=end, length=None, matches=None,
			mismatches=None, gaps=None, percent_identity=None)

	def _find_cdr3(self):
		"""
		Return a repaired AlignmentSummary object that describes the CDR3 region.
		Return None if no CDR3 detected.
		"""
		if 'V' not in self.hits or 'J' not in self.hits:
			return None
		if self.chain not in self.CHAINS:
			return None

		# CDR3 start
		cdr3_ref_start = self._database.v_cdr3_start(self.hits['V'].subject_id, self.CHAINS[self.chain])
		if cdr3_ref_start is None:
			return None
		cdr3_query_start = self.hits['V'].query_position(reference_position=cdr3_ref_start)
		if cdr3_query_start is None:
			# Alignment is not long enough to cover CDR3 start position; try to rescue it
			# by assuming that the alignment would continue without indels.
			hit = self.hits['V']
			cdr3_query_start = hit.query_end + (cdr3_ref_start - hit.subject_end)

		# CDR3 end
		cdr3_ref_end = self._database.j_cdr3_end(self.hits['J'].subject_id, self.CHAINS[self.chain])
		if cdr3_ref_end is None:
			return None

		cdr3_query_end = self.hits['J'].query_position(reference_position=cdr3_ref_end)
		if cdr3_query_end is None:
			return None

		return AlignmentSummary(start=cdr3_query_start, stop=cdr3_query_end, length=None, matches=None,
			mismatches=None, gaps=None, percent_identity=None)

	def _fixed_v_hit(self):
		"""
		Extend the V hit to the left if it does not start at the first nucleotide of the V gene.
		"""
		hit = self.hits['V']
		d = hit._asdict()
		while d['subject_start'] > 0 and d['query_start'] > 0:
			d['query_start'] -= 1
			d['subject_start'] -= 1
			preceding_query_base = self.full_sequence[d['query_start']]
			d['query_alignment'] = preceding_query_base + d['query_alignment']
			if self._database.v:
				reference = self._database.v[hit.subject_id]
				preceding_base = reference[d['subject_start']]
			else:
				preceding_base = 'N'
			d['subject_alignment'] = preceding_base + d['subject_alignment']
		return Hit(**d)

	def _aa_region_mutations(self, region, aa_sequence):
		"""
		For a given region, compute number of amino-acid level mutations
		between the query and reference.

		Return a tuple no_of_mutations, aa_region_length. Raise KeyError if no
		information for the region is available.
		"""
		reference_sequence = self._database.v_regions_aa[self.v_gene][region]
		return edit_distance(aa_sequence, reference_sequence), len(reference_sequence)

	def asdict(self):
		"""
		Return a flattened representation of this record as a dictionary.
		The dictionary can then be used with e.g. a csv.DictWriter or
		pandas.DataFrame.from_items.
		"""
		nt_regions = dict()
		aa_regions = dict()
		for region in ('FR1', 'CDR1', 'FR2', 'CDR2', 'FR3', 'CDR3'):
			nt_seq = self.region_sequence(region)
			nt_regions[region] = nt_seq
			aa_regions[region] = nt_to_aa(nt_seq) if nt_seq else None

		vdj_nt = self.vdj_sequence
		vdj_aa = nt_to_aa(vdj_nt) if vdj_nt else None

		def nt_mutation_rate(region):
			"""Nucleotide-level mutation rate in percent"""
			if region in self.alignments:
				rar = self.alignments[region]
				if rar is None or rar.percent_identity is None:
					return None
				return 100. - rar.percent_identity
			else:
				return None

		def j_aa_mutation_rate():
			if 'J' not in self.hits:
				return None
			j_subject_id = self.hits['J'].subject_id
			if self.chain not in self.CHAINS:
				return None
			cdr3_ref_end = self._database.j_cdr3_end(j_subject_id, self.CHAINS[self.chain])
			if cdr3_ref_end is None:
				return None
			cdr3_query_end = self.hits['J'].query_position(reference_position=cdr3_ref_end)
			if cdr3_query_end is None:
				return None

			query = self.full_sequence[cdr3_query_end:self.hits['J'].query_end]
			try:
				query_aa = nt_to_aa(query)
			except ValueError:
				return None
			ref = self._database.j[j_subject_id][cdr3_ref_end:self.hits['J'].subject_end]
			try:
				ref_aa = nt_to_aa(ref)
			except ValueError:
				return None
			if not ref_aa:
				return None
			return 100. * edit_distance(ref_aa, query_aa) / len(ref_aa)

		def aa_mutation_rates():
			"""Amino-acid level mutation rates for all regions and V in percent"""
			rates = dict()
			rates['J'] = j_aa_mutation_rate()
			v_aa_mutations = 0
			v_aa_length = 0
			for region in ('FR1', 'CDR1', 'FR2', 'CDR2', 'FR3'):
				if not aa_regions[region]:
					break
				try:
					mutations, length = self._aa_region_mutations(region, aa_regions[region])
				except KeyError:
					break
				rates[region] = 100. * mutations / length
				v_aa_mutations += mutations
				v_aa_length += length
			else:
				rates['V'] = 100. * v_aa_mutations / v_aa_length
				return rates
			return dict(FR1=None, CDR1=None, FR2=None, CDR2=None, FR3=None, V=None, J=None)

		aa_rates = aa_mutation_rates()

		if 'V' in self.hits:
			v_nt = self.hits['V'].query_sequence
			v_aa = nt_to_aa(v_nt)
			v_shm = 100. - self.hits['V'].percent_identity
			v_errors = self.hits['V'].errors
			v_covered = 100. * self.hits['V'].covered()
			v_evalue = self.hits['V'].evalue
		else:
			v_nt = None
			v_aa = None
			v_shm = None
			v_errors = None
			v_covered = None
			v_evalue = None
		if 'D' in self.hits:
			d_errors = self.hits['D'].errors
			d_covered = 100. * self.hits['D'].covered()
			d_evalue = self.hits['D'].evalue
		else:
			d_errors = None
			d_covered = None
			d_evalue = None
		if 'J' in self.hits:
			j_nt = self.hits['J'].query_sequence
			j_shm = 100. - self.hits['J'].percent_identity
			j_errors = self.hits['J'].errors
			j_covered = 100. * self.hits['J'].covered()
			j_evalue = self.hits['J'].evalue
		else:
			j_nt = None
			j_shm = None
			j_errors = None
			j_covered = None
			j_evalue = None
		v_end = getattr(self.junction, 'v_end', None)
		vd_junction = getattr(self.junction, 'vd_junction', None)
		d_region = getattr(self.junction, 'd_region', None)
		dj_junction = getattr(self.junction, 'dj_junction', None)

		return dict(
			count=self.size,
			V_gene=self.v_gene,
			D_gene=self.d_gene,
			J_gene=self.j_gene,
			chain=self.chain,
			stop=self.has_stop,
			V_covered=v_covered,
			D_covered=d_covered,
			J_covered=j_covered,
			V_evalue=v_evalue,
			D_evalue=d_evalue,
			J_evalue=j_evalue,
			FR1_SHM=nt_mutation_rate('FR1'),
			CDR1_SHM=nt_mutation_rate('CDR1'),
			FR2_SHM=nt_mutation_rate('FR2'),
			CDR2_SHM=nt_mutation_rate('CDR2'),
			FR3_SHM=nt_mutation_rate('FR3'),
			V_SHM=v_shm,
			J_SHM=j_shm,
			V_aa_mut=aa_rates['V'],
			J_aa_mut=aa_rates['J'],
			FR1_aa_mut=aa_rates['FR1'],
			CDR1_aa_mut=aa_rates['CDR1'],
			FR2_aa_mut=aa_rates['FR2'],
			CDR2_aa_mut=aa_rates['CDR2'],
			FR3_aa_mut=aa_rates['FR3'],
			V_errors=v_errors,
			D_errors=d_errors,
			J_errors=j_errors,
			UTR=self.utr,
			leader=self.leader,
			CDR1_nt=nt_regions['CDR1'],
			CDR1_aa=aa_regions['CDR1'],
			CDR2_nt=nt_regions['CDR2'],
			CDR2_aa=aa_regions['CDR2'],
			CDR3_nt=nt_regions['CDR3'],
			CDR3_aa=aa_regions['CDR3'],
			V_nt=v_nt,
			V_aa=v_aa,
			V_end=v_end,
			V_CDR3_start=self.v_cdr3_start,
			VD_junction=vd_junction,
			D_region=d_region,
			DJ_junction=dj_junction,
			J_nt=j_nt,
			VDJ_nt=vdj_nt,
			VDJ_aa=vdj_aa,
			name=self.query_name,
			barcode=self.barcode,
			genomic_sequence=self.genomic_sequence,
		)


class ParseError(Exception):
	pass


class IgBlastParser:
	"""
	Parser for IgBLAST results. Works only when IgBLAST was run with
	the option -outfmt "7 sseqid qstart qseq sstart sseq pident slen".
	"""
	BOOL = {'Yes': True, 'No': False, 'N/A': None}
	FRAME = {'In-frame': True, 'Out-of-frame': False, 'N/A': None}
	SECTIONS = frozenset([
		'# Query:',
		'# V-(D)-J rearrangement summary',
		'# V-(D)-J junction details',
		'# Alignment summary',
		'# Hit table',
		'Total queries = ',
	])

	def __init__(self, sequences, igblast_lines):
		self._sequences = sequences
		self._igblast_lines = igblast_lines

	def __iter__(self):
		"""
		Yield IgBlastRecord objects
		"""
		zipped = zip(self._sequences, split_by_section(self._igblast_lines, ['# IGBLASTN']))
		for fasta_record, (record_header, record_lines) in zipped:
			# 'IGBLASTN 2.2.29+': IgBLAST 1.4.0
			# 'IGBLASTN 2.3.1+': IgBLAST 1.5.0
			# 'IGBLASTN 2.5.1+': IgBLAST 1.6.1
			# 'IGBLASTN 2.6.1+': IgBLAST 1.7.0
			assert record_header in {'# IGBLASTN 2.2.29+', '# IGBLASTN 2.3.1+', '# IGBLASTN 2.6.1+'}
			yield self._parse_record(record_lines, fasta_record)

	def _parse_record(self, record_lines, fasta_record):
		"""
		Parse a single IgBLAST record
		"""
		hits = dict()
		# All of the sections are optional, so we need to set default values here.
		query_name = None
		junction = None
		v_gene, d_gene, j_gene, chain, has_stop, in_frame, is_productive, strand = [None] * 8
		alignments = dict()
		for section, lines in split_by_section(record_lines, self.SECTIONS):
			if section.startswith('# Query: '):
				query_name = section.split(': ')[1]
			elif section.startswith('# V-(D)-J rearrangement summary'):
				fields = lines[0].split('\t')
				if len(fields) == 7:
					# No D assignment
					v_gene, j_gene, chain, has_stop, in_frame, is_productive, strand = fields
					d_gene = None
				else:
					v_gene, d_gene, j_gene, chain, has_stop, in_frame, is_productive, strand = fields
				v_gene = none_if_na(v_gene)
				d_gene = none_if_na(d_gene)
				j_gene = none_if_na(j_gene)
				chain = none_if_na(chain)
				has_stop = self.BOOL[has_stop]
				in_frame = self.FRAME[in_frame]
				is_productive = self.BOOL[is_productive]
				strand = strand if strand in '+-' else None
			elif section.startswith('# V-(D)-J junction details'):
				fields = lines[0].split('\t')
				if len(fields) == 5:
					junction = JunctionVDJ(
						v_end=fields[0],
						vd_junction=fields[1],
						d_region=fields[2],
						dj_junction=fields[3],
						j_start=fields[4]
					)
				else:
					junction = JunctionVJ(
						v_end=fields[0],
						vj_junction=fields[1],
						j_start=fields[2])
			elif section.startswith('# Alignment summary'):
				for line in lines:
					fields = line.split('\t')
					if len(fields) == 8 and fields[0] != 'Total':
						summary = self._parse_alignment_summary(fields[1:])
						region_name, _, imgt = fields[0].partition('-')
						assert imgt in ('IMGT', 'IMGT (germline)')
						alignments[region_name] = summary
			elif section.startswith('# Hit table'):
				for line in lines:
					if not line or line.startswith('#'):
						continue
					hit, gene = self._parse_hit(line)
					assert gene in ('V', 'D', 'J')
					assert gene not in hits, "Two hits for same gene found"
					hits[gene] = hit
			elif section.startswith('Total queries = '):
				continue

		assert fasta_record.name == query_name
		full_sequence = fasta_record.sequence.upper()
		if strand == '-':
			full_sequence = reverse_complement(full_sequence)

		if __debug__:
			for gene in ('V', 'D', 'J'):
				if gene not in hits:
					continue
				hit = hits[gene]
				qsequence = hit.query_sequence

				# IgBLAST removes the trailing semicolon (why, oh why??)
				qname = query_name[:-1] if query_name.endswith(';') else query_name
				assert chain in (None, 'VL', 'VH', 'VK', 'NON', 'VA', 'VB', 'VG', 'VD'), chain
				assert qsequence == full_sequence[hit.query_start:hit.query_start+len(qsequence)]

		return IgBlastRecord(
			query_name=query_name,
			alignments=alignments,
			v_gene=v_gene,
			d_gene=d_gene,
			j_gene=j_gene,
			chain=chain,
			has_stop=has_stop,
			in_frame=in_frame,
			is_productive=is_productive,
			strand=strand,
			hits=hits,
			full_sequence=full_sequence,
			junction=junction)

	def _parse_alignment_summary(self, fields):
		start, stop, length, matches, mismatches, gaps = (int(v) for v in fields[:6])
		percent_identity = float(fields[6])
		return AlignmentSummary(
			start=start - 1,
			stop=stop,
			length=length,
			matches=matches,
			mismatches=mismatches,
			gaps=gaps,
			percent_identity=percent_identity
		)

	def _parse_hit(self, line):
		"""
		Parse a line of the "Hit table" section and return a tuple (hit, gene)
		where hit is a Hit object.
		"""
		(gene, subject_id, query_start, query_alignment, subject_start, subject_alignment,
			percent_identity, subject_length, evalue) = line.split('\t')
		query_start = int(query_start) - 1
		subject_start = int(subject_start) - 1
		subject_length = int(subject_length)  # Length of original subject sequence
		# Percent identity is calculated by IgBLAST as
		# 100 - errors / alignment_length and then rounded to two decimal digits
		percent_identity = float(percent_identity)
		evalue = float(evalue)
		hit = Hit(subject_id, query_start, query_alignment, subject_start,
			subject_alignment, subject_length, percent_identity, evalue)
		return hit, gene


class ExtendedIgBlastParser:
	def __init__(self, sequences, igblast_lines, database):
		self._parser = IgBlastParser(sequences, igblast_lines)
		self._database = database

	def __iter__(self):
		for record in self._parser:
			yield ExtendedIgBlastRecord(self._database, **record.__dict__)


class TableWriter:
	def __init__(self, file):
		self._file = file
		self._writer = csv.DictWriter(file, fieldnames=ExtendedIgBlastRecord.columns, delimiter='\t')
		self._writer.writeheader()

	@staticmethod
	def yesno(v):
		"""
		Return "yes", "no" or None for boolean value v, which may also be None.
		"""
		if v is None:
			return None
		return ["no", "yes"][v]

	def write(self, d):
		"""
		Write the IgBLAST record (must be given as dictionary) to the output
		file.
		"""
		d = d.copy()
		d['stop'] = self.yesno(d['stop'])
		for name in ('V_covered', 'D_covered', 'J_covered',
				'FR1_SHM', 'CDR1_SHM', 'FR2_SHM', 'CDR2_SHM', 'FR3_SHM',
				'V_SHM', 'J_SHM', 'V_aa_mut', 'J_aa_mut',
				'FR1_aa_mut', 'CDR1_aa_mut', 'FR2_aa_mut', 'CDR2_aa_mut', 'FR3_aa_mut'):
			if d[name] is not None:
				d[name] = '{:.1f}'.format(d[name])
		for name in ('V_evalue', 'D_evalue', 'J_evalue'):
			if d[name] is not None:
				d[name] = '{:G}'.format(d[name])
		self._writer.writerow(d)



"""
Test the SiblingMerger class
"""
import pandas as pd
import pytest

from igdiscover.discover import SiblingMerger, SiblingInfo
from igdiscover.germlinefilter import SequenceMerger, SequenceInfo
from igdiscover.utils import UniqueNamer
from igdiscover.rename import PrefixDict


def test_0():
	merger = SiblingMerger()
	assert list(merger) == []


def test_1():
	merger = SiblingMerger()
	group = pd.DataFrame([1, 2, 3])
	info = SiblingInfo('ACCGGT', False, 'name1', group)
	merger.add(info)
	assert list(merger) == [info]


def test_2():
	merger = SiblingMerger()
	group = pd.DataFrame([1, 2, 3])
	merger.add(SiblingInfo('ACCGGT', False, 'name1', group))
	merger.add(SiblingInfo('ACCGGT', False, 'name2', group))
	sisters = list(merger)
	assert len(sisters) == 1
	assert sisters[0].sequence == 'ACCGGT'
	assert not sisters[0].requested
	assert sisters[0].name == 'name1;name2'


def test_requested():
	merger = SiblingMerger()
	group = pd.DataFrame([1, 2, 3])
	merger.add(SiblingInfo('ACCGGT', True, 'name1', group))
	merger.add(SiblingInfo('ACCGGT', False, 'name2', group))
	sisters = list(merger)
	assert sisters[0].requested


def test_prefix():
	merger = SiblingMerger()
	group = pd.DataFrame([1, 2, 3])
	merger.add(SiblingInfo('ACCGGTAACGT', True, 'name1', group))
	merger.add(SiblingInfo('ACCGGT', False, 'name2', group))
	info2 = SiblingInfo('TGATACC', False, 'name3', group)
	merger.add(info2)
	sisters = list(merger)
	assert len(sisters) == 2
	assert sisters[0].sequence == 'ACCGGTAACGT'
	assert sisters[0].name == 'name1;name2'
	assert sisters[1] == info2


def test_with_N():
	merger = SiblingMerger()
	group = pd.DataFrame([1, 2, 3])
	merger.add(SiblingInfo('ACCNGTAANGT', True, 'name1', group))
	merger.add(SiblingInfo('ANCGGT', False, 'name2', group))
	info2 = SiblingInfo('TGATACC', False, 'name3', group)
	merger.add(info2)
	sisters = list(merger)
	assert len(sisters) == 2
	assert sisters[0].sequence == 'ACCGGTAANGT'
	assert sisters[0].name == 'name1;name2'
	assert sisters[1] == info2


def test_unique_namer():
	namer = UniqueNamer()
	assert namer('Name') == 'Name'
	assert namer('AnotherName') == 'AnotherName'
	assert namer('Name') == 'NameA'
	assert namer('Name') == 'NameB'
	assert namer('YetAnotherName') == 'YetAnotherName'
	assert namer('Name') == 'NameC'
	assert namer('NameC') == 'NameCA'


def SI(sequence, name, clonotypes, whitelisted=False):
	return SequenceInfo(
		sequence=sequence,
		name=name,
		clonotypes=clonotypes,
		exact=100,
		Ds_exact=10,
		cluster_size=100,
		whitelisted=whitelisted,
		is_database=False,
		cluster_size_is_accurate=True,
		CDR3_start=len(sequence),
		row=None
	)


def test_sequence_merger_withCDR3():
	merger = SequenceMerger(max_differences=1, cross_mapping_ratio=0, clonotype_ratio=None,
		exact_ratio=None, unique_d_ratio=None, unique_d_threshold=10)
	infos = [
		SI('ACGTTA', 'Name1', 15),
		SI('ACGTTAT', 'Name2', 100),  # kept because it is longer
		SI('ACGCCAT', 'Name3', 15),   # kept because edit distance > 1
		SI('ACGGTAT', 'Name5', 120),  # kept because it has more clonotypes
	]
	merger.add(infos[0]); merged = list(merger)
	assert len(merged) == 1 and merged[0] == infos[0]
	merger.add(infos[0]); merged = list(merger)
	assert len(merged) == 1 and merged[0] == infos[0]
	merger.add(infos[1]); merged = list(merger)
	assert len(merged) == 1 and merged[0] == infos[1]
	merger.add(infos[2]); merged = list(merger)
	assert len(merged) == 2 and merged[0] == infos[1] and merged[1] == infos[2]
	merger.add(infos[3]); merged = list(merger)
	assert len(merged) == 2 and merged[0] == infos[2] and merged[1] == infos[3]


def test_sequence_merger_prefix():
	merger = SequenceMerger(max_differences=1, cross_mapping_ratio=0, clonotype_ratio=None,
		exact_ratio=None,
		unique_d_ratio=None, unique_d_threshold=10)
	infos = [
		SI('AAATAA', 'Name1', 117),
		SI('AAACAAG', 'Name2', 10),
	]
	merger.add(infos[0])
	merger.add(infos[1])
	merged = list(merger)
	assert len(merged) == 1
	assert merged[0] == infos[0], (merged, infos)


def test_merger_check_all_previous():
	merger = SequenceMerger(max_differences=1, cross_mapping_ratio=0, clonotype_ratio=None,
		exact_ratio=None,
		unique_d_ratio=None, unique_d_threshold=10)
	infos = [
		SI('ATAAAA', 's1', 11),
		SI('AAGAAA', 's2', 12),
		SI('AAACAA', 's3', 13),
		SI('AAAAAA', 'Final', 200),
	]
	for info in infos:
		merger.add(info)
	merged = list(merger)
	assert len(merged) == 1
	assert merged[0] == infos[3]


class TestPrefixDict:
	def setup(self):
		self.pd = PrefixDict([
			('AAACCT', 7),
			('AGAAA', 11),
			('AGAAC', 13)
		])

	def test_ambiguous(self):
		with pytest.raises(KeyError):
			self.pd['AGAA']

	def test_missing(self):
		with pytest.raises(KeyError):
			self.pd['TTAAG']

	def test_existing(self):
		assert self.pd['AAACCT'] == 7
		assert self.pd['AAACCT'] == 7
		assert self.pd['AAA'] == 7
		assert self.pd['AGAAAT'] == 11
		assert self.pd['AGAACT'] == 13



"""
Find common V genes between two different antibody libraries.
"""
import logging
from collections import Counter

from .table import read_table

logger = logging.getLogger(__name__)


def add_arguments(parser):
	arg = parser.add_argument
	arg('--minimum-frequency', '-n', type=int, metavar='N',
		default=None,
		help='Minimum number of datasets in which sequence must occur (default is no. of files divided by two)')
	arg('table', help='Tables with parsed and filtered IgBLAST results (give at least two)', nargs='+')


def main(args):
	if args.minimum_frequency is None:
		# args.table is a list of file names
		minimum_frequency = max((len(args.table) + 1) // 2, 2)
	else:
		minimum_frequency = args.minimum_frequency
	logger.info('Minimum frequency set to %s', minimum_frequency)

	# Read in tables
	tables = []
	for path in args.table:
		table = read_table(path)
		table = table.loc[:,['V_gene', 'V_SHM', 'V_nt', 'name']]
		tables.append(table)

	# Count V sequence occurrences
	counter = Counter()
	for table in tables:
		counter.update(set(table.V_nt))

	# Find most frequent occurrences and print result
	print('Frequency', 'Gene', '%SHM', 'Sequence', sep='\t')
	for sequence, frequency in counter.most_common():
		if frequency < minimum_frequency:
			break
		names = []
		gene = None
		for table in tables:
			matching_rows = table[table.V_nt == sequence]
			if matching_rows.empty:
				continue
			names.extend(matching_rows.name)
			if gene is None:
				row = matching_rows.iloc[0]
				gene = row['V_gene']
				shm = row['V_SHM']
		print(frequency, gene, shm, sequence, *names, sep='\t')



"""
Group sequences that share a barcode (molecular identifier, MID)

Since the same barcode can sometimes be used by different sequences, the CDR3
sequence can further be used to distinguish sequences. You can choose between
using either a 'pseudo CDR3' sequence, which encompasses by default bases 80
to 61 counted from the 3' end. Or you can use the real CDR3 detected with a
regular expression.

If grouping by CDR3s is enabled, sequences with identical barcode and CDR3
must additionally have a similar length. If the length differs by more than
2 bp, they are put into different groups.

The barcode can be in the 5' end or the 3' end of the sequence.

Use --trim-g to remove initial runs of G at the 5' end (artifact from RACE protocol).
These are removed after the barcode is removed.

For all the found groups, one sequence is output to standard output (in FASTA
format). Which sequence that is depends on the group size:
- If the group consists of a single sequence, that sequence is output
- If the group consists of two sequences, one of them is picked randomly
- If the group has at least three sequences, a consensus is computed. The
  consensus is output if it contains no ambiguous bases. Otherwise, also here a
  random sequence is chosen.
"""
# NOTES
#
# - Different lengths of the initial G run cannot be used to distinguish sequences
#  since they can come from polymerase problems in homopolymers.
# - There are some indels in homopolymers (probably PCR problem)
# - There are also regular sequencing errors in the initial run of G nucleotides.
# - Some paired reads aren’t correctly merged into single reads. They end up being
#   too long.
# - When grouping by barcode and pseudo CDR3, sequence lengths vary within groups.
#   However, this affects only ~1% of sequences, so it is not necessary to compute
#   a multiple alignment. Just taking the consensus will drown the incorrect
#   sequences, at least if the group size is large.
# - It does not hurt to reduce the minimimum number of sequences per group for
#   taking a consensus to 2, but it also does not help much (results in 0.5% more
#   sequences). (The consensus is only successful if both sequences are identical.)
#   However, since this is also a simple way to deal with exact duplicates, we do
#   it anyway and can then skip the separate duplicate removal step (VSEARCH).

# TODO
# - Use pandas.DataFrame

import csv
import sys
import logging
from collections import Counter, defaultdict
from contextlib import ExitStack
from itertools import islice

import matplotlib
matplotlib.use('agg')
import matplotlib.pyplot as plt
from sqt.align import consensus
from sqt import SequenceReader
from xopen import xopen
from .species import find_cdr3
from .cluster import Graph
from .utils import slice_arg

# minimum number of sequences needed for attempting to compute a consensus
MIN_CONSENSUS_SEQUENCES = 3


logger = logging.getLogger(__name__)


def add_arguments(parser):
	arg = parser.add_argument
	group = parser.add_mutually_exclusive_group()
	group.add_argument('--real-cdr3', action='store_true', default=False,
		help='In addition to barcode, group sequences by real CDR3 (detected with regex).')
	group.add_argument('--pseudo-cdr3', nargs='?', default=None,
		type=slice_arg, const=slice(-80, -60), metavar='START:END',
		help='In addition to barcode, group sequences by pseudo CDR3. '
			'If START:END is omitted, use -80:-60.')
	arg('--groups-output', metavar='FILE', default=None,
		help='Write tab-separated table with groups to FILE')
	arg('--plot-sizes', metavar='FILE', default=None,
		help='Plot group sizes to FILE (.png or .pdf)')
	arg('--limit', default=None, type=int, metavar='N',
		help='Limit processing to the first N reads')
	arg('--trim-g', action='store_true', default=False,
		help="Trim 'G' nucleotides at 5' end")
	arg('--minimum-length', '-l', type=int, default=0,
		help='Minimum sequence length')
	arg('--barcode-length', '-b', type=int, default=12,
		help="Length of barcode. Positive for 5' barcode, negative for 3' barcode. Default: %(default)s")
	arg('fastx', metavar='FASTA/FASTQ',
		help='FASTA or FASTQ file (can be gzip-compressed) with sequences')


def hamming_neighbors(s):
	"""Return sequences that are at hamming distance 1 and return also s itself"""
	for i in range(len(s)):
		for c in 'ACGT':
			if s[i] != c:
				yield s[:i] + c + s[i+1:]
	yield s


def cluster_sequences(records):
	"""
	Single-linkage clustering. Two sequences are linked if
	- their (pseudo-) CDR3 sequences have a hamming distance of at most 1
	- and their lengths differs by at most 2.
	"""
	if len(records) == 1:  # TODO check if this helps
		return [records]

	# Cluster unique CDR3s first
	cdr3s = set(r.cdr3 for r in records)
	sorted_cdr3s = sorted(cdr3s)  # For reproducibility
	graph = Graph(sorted_cdr3s)
	for cdr3 in sorted_cdr3s:
		for neighbor in hamming_neighbors(cdr3):
			if neighbor in cdr3s:
				graph.add_edge(cdr3, neighbor)
	cdr3_components = graph.connected_components()

	# Maps CDR3 sequence to list of records of sequence that have that CDR3
	cdr3_records = defaultdict(list)
	for r in records:
		cdr3_records[r.cdr3].append(r)

	components = []
	for cdr3_component in cdr3_components:
		component_records = []
		for cdr3 in cdr3_component:
			component_records.extend(cdr3_records[cdr3])

		component_records.sort(key=lambda r: len(r.sequence))
		component = []
		prev_length = None
		for r in component_records:
			l = len(r.sequence)
			if prev_length is not None and l > prev_length + 2:
				# Start a new component
				components.append(component)
				component = []
			component.append(r)
			prev_length = l
		if component:
			components.append(component)

	assert sum(len(component) for component in components) == len(records)
	assert all(components)  # Components must be non-empty
	return components


GROUPS_HEADER = ['barcode', 'cdr3', 'name', 'sequence']


def write_group(csvfile, barcode, sequences):
	for sequence in sequences:
		row = [barcode, sequence.cdr3, sequence.name.split(maxsplit=1)[0], sequence.sequence]
		csvfile.writerow(row)
	csvfile.writerow([])


def collect_barcode_groups(
		fastx, barcode_length, trim_g, limit, minimum_length, pseudo_cdr3, real_cdr3):
	"""
	fastx -- path to FASTA or FASTQ input

	"""
	group_by_cdr3 = pseudo_cdr3 or real_cdr3
	if group_by_cdr3:
		cdr3s = set()
	# Map barcodes to lists of sequences
	barcodes = defaultdict(list)
	n = 0
	too_short = 0
	regex_fail = 0
	with SequenceReader(fastx) as f:
		for record in islice(f, 0, limit):
			if len(record) < minimum_length:
				too_short += 1
				continue
			if barcode_length > 0:
				barcode = record.sequence[:barcode_length]
				unbarcoded = record[barcode_length:]
			else:
				barcode = record.sequence[barcode_length:]
				unbarcoded = record[:barcode_length]

			if trim_g:
				# The RACE protocol leads to a run of non-template Gs in the beginning
				# of the sequence, after the barcode.
				unbarcoded.sequence = unbarcoded.sequence.lstrip('G')
				if unbarcoded.qualities:
					unbarcoded.qualities = unbarcoded.qualities[-len(unbarcoded.sequence):]

			if real_cdr3:
				match = find_cdr3(unbarcoded.sequence, chain='VH')
				if match:
					cdr3 = unbarcoded.sequence[match[0]:match[1]]
				else:
					regex_fail += 1
					continue
			elif pseudo_cdr3:
				cdr3 = unbarcoded.sequence[pseudo_cdr3]
			if group_by_cdr3:
				unbarcoded.cdr3 = cdr3  # TODO slight abuse of Sequence objects
				cdr3s.add(cdr3)
			barcodes[barcode].append(unbarcoded)
			n += 1

	logger.info('%s sequences in input', n + too_short + regex_fail)
	logger.info('%s sequences long enough', n + regex_fail)
	if real_cdr3:
		logger.info('Using the real CDR3')
		logger.info('%s times (%.2f%%), the CDR3 regex matched', n, n / (n + regex_fail) * 100)
	elif pseudo_cdr3:
		logger.info('Using the pseudo CDR3')
	if group_by_cdr3:
		logger.info('%s unique CDR3s', len(cdr3s))

	return barcodes


def main(args):
	if args.barcode_length == 0:
		sys.exit("Barcode length must be non-zero")

	group_by_cdr3 = args.pseudo_cdr3 or args.real_cdr3
	barcodes = collect_barcode_groups(args.fastx, args.barcode_length, args.trim_g,
		args.limit, args.minimum_length, args.pseudo_cdr3, args.real_cdr3)

	logger.info('%s unique barcodes', len(barcodes))
	barcode_singletons = sum(1 for seqs in barcodes.values() if len(seqs) == 1)
	logger.info('%s barcodes used by only a single sequence (singletons)', barcode_singletons)

	with ExitStack() as stack:
		if args.groups_output:
			group_out = csv.writer(stack.enter_context(
				xopen(args.groups_output, 'w')), delimiter='\t', lineterminator='\n')
			group_out.writerow(GROUPS_HEADER)
		else:
			group_out = None
		too_few = 0
		n_clusters = 0
		n_singletons = 0
		n_consensus = 0
		n_ambiguous = 0
		sizes = []
		for barcode in sorted(barcodes):
			sequences = barcodes[barcode]
			if len(sequences) != len(set(s.name for s in sequences)):
				logger.error('Duplicate sequence records detected')
				sys.exit(1)
			if group_by_cdr3:
				clusters = cluster_sequences(sequences)  # it’s a list of lists
			else:
				# TODO it would be useful to do the clustering by length that cluster_sequences() does
				clusters = [sequences]
			n_clusters += len(clusters)
			for cluster in clusters:
				sizes.append(len(cluster))
				if group_out:
					write_group(group_out, barcode, cluster)
				if len(cluster) == 1:
					n_singletons += 1
				if len(cluster) < MIN_CONSENSUS_SEQUENCES:
					too_few += 1
					sequence = cluster[0].sequence
					name = cluster[0].name
					if group_by_cdr3:
						cdr3 = cluster[0].cdr3
				else:
					cons = consensus({s.name: s.sequence for s in cluster}, threshold=0.501)
					if 'N' in cons:
						# Pick the first sequence as the output sequence
						sequence = cluster[0].sequence
						name = cluster[0].name
						if group_by_cdr3:
							cdr3 = cluster[0].cdr3
						n_ambiguous += 1
					else:
						sequence = cons
						n_consensus += 1
						if group_by_cdr3:
							cdr3 = Counter(cl.cdr3 for cl in cluster).most_common(1)[0][0]
						name = 'consensus{}'.format(n_consensus)

				name = name.split(maxsplit=1)[0]
				if name.endswith(';'):
					name = name[:-1]
				if group_by_cdr3:
					print('>{};barcode={};cdr3={};size={};\n{}'.format(name, barcode,
						cdr3, len(cluster), sequence))
				else:
					print('>{};barcode={};size={};\n{}'.format(name, barcode,
						len(cluster), sequence))

	logger.info('%d clusters (%d singletons)', n_clusters, n_singletons)
	logger.info('%d consensus sequences computed (from groups that had at least %d sequences)',
		n_consensus + n_ambiguous, MIN_CONSENSUS_SEQUENCES)
	logger.info('%d of those had no ambiguous bases', n_consensus)
	if args.groups_output:
		logger.info('Groups written to %r', args.groups_output)

	if args.plot_sizes:
		matplotlib.rcParams.update({'font.size': 14})
		fig = plt.figure()
		ax = fig.gca()
		v, _, _ = ax.hist(sizes, bins=100)
		ax.set_ylim(0, v[1:].max()*1.1)
		ax.set_xlabel('Group size')
		ax.set_ylabel('Read frequency')
		ax.set_title('Histogram of group sizes (>1)')
		ax.grid(axis='x')
		ax.tick_params(direction='out', top=False, right=False)
		fig.set_tight_layout(True)
		fig.savefig(args.plot_sizes)
		logger.info('Plotted group sizes to %r', args.plot_sizes)




"""
Rename and reorder records in a FASTA file

Sequences can be renamed according to the sequences in a template file.
The sequences in the target file will get the name that they have in the
template file. Sequences are considered to be equivalent if one is a prefix of
the other.

Sequences can also be ordered by name, either alphabetically or by
the order given in a template file. For comparison, only the 'gene part'
of the name is used (before the '*'). For example, for 'VH4-4*01', the name 'VH4-4'
is looked up in the template. Alphabetic order is the default. Use --no-sort
to disable sorting entirely.
"""
import sys
import logging
from sqt import FastaReader
from .utils import natural_sort_key

logger = logging.getLogger(__name__)


def add_arguments(parser):
	arg = parser.add_argument
	arg('--no-sort', dest='sort', action='store_false', default=True,
		help='Do not sort sequences by name')
	arg('--not-found', metavar='TEXT', default=' (not found)',
		help='Append this text to the record name when the sequence was not found '
		'in the template. Default: %(default)r')
	arg('--rename-from', metavar='TEMPLATE',
		help='FASTA template file with correctly named sequences. If a sequence in '
			'the target file is identical to one in the template, it is assigned the '
			'name of the sequence in the template.')
	arg('--order-by', metavar='TEMPLATE',
		help='FASTA template that contains genes in the desired order. '
			'If a name contains a "*" (asterisk), only the preceding part '
			'is used. Thus, VH4-4*01 and VH4-4*02 are equivalent.')
	arg('target', help='FASTA file with to-be renamed sequences')


class PrefixDict:
	"""
	A dict that maps strings to values, but where a prefix of a key is enough
	to retrieve the value.
	"""
	def __init__(self, items):
		self._items = []
		for k, v in items:
			self.add(k, v)

	def add(self, k, v):
		try:
			self[k]
			raise ValueError('Key {!r} already exists'.format(k))
		except KeyError:
			self._items.append((k, v))

	def __getitem__(self, key):
		found = None
		for seq, value in self._items:
			if seq.startswith(key) or key.startswith(seq):
				if found is not None:
					# TODO don't use keyerror here
					raise KeyError('Key {!r} is ambiguous'.format(key))
				found = value
		if found is None:
			raise KeyError(key)
		return found

	def get(self, key, default=None):
		try:
			v = self[key]
		except KeyError:
			return default
		else:
			return v

	def __len__(self):
		return len(self._items)


class GeneMissing(Exception):
	pass


def gene_name(record):
	record_name, _, record_comment = record.name.partition(' ')
	return record_name.partition('*')[0]


def sorted_by_gene(records, gene_order):
	d = {name: i for i, name in enumerate(gene_order)}

	def keyfunc(record):
		gene = gene_name(record)
		try:
			return d[gene]
		except KeyError:
			raise GeneMissing(gene)

	return sorted(records, key=keyfunc)


def main(args):
	if args.rename_from:
		with FastaReader(args.rename_from) as fr:
			template = PrefixDict([])
			for record in fr:
				try:
					template.add(record.sequence.upper(), record.name)
				except ValueError:
					logger.error('Sequences in entry %r and %r are duplicate',
						record.name, template[record.sequence.upper()])
		logger.info('Read %d entries from template', len(template))
	else:
		template = None

	if args.order_by:
		with FastaReader(args.order_by) as fr:
			gene_order = [gene_name(r) for r in fr]
	else:
		gene_order = None

	with FastaReader(args.target) as fr:
		sequences = list(fr)

	# Rename
	renamed = 0
	if template is not None:
		for record in sequences:
			name = template.get(record.sequence.upper())
			if name is None:
				name = record.name + args.not_found
			else:
				renamed += 1
			# Replace record’s name, leaving comment intact
			record_name, _, record_comment = record.name.partition(' ')
			if record_comment:
				record.name = name + ' ' + record_comment
			else:
				record.name = name

	# Reorder
	if gene_order:
		try:
			sequences = sorted_by_gene(sequences, gene_order)
		except GeneMissing as e:
			logger.error('Gene "%s" not found in the --order-by template file', e)
			sys.exit(1)
	elif args.sort:
		sequences = sorted(sequences, key=lambda s: natural_sort_key(s.name))
	for record in sequences:
		print('>{}\n{}'.format(record.name, record.sequence))
	logger.info('Wrote %s FASTA records (%d sequences found in template)', len(sequences), renamed)



"""
Run IgBLAST and output a result table

This is a wrapper for the "igblastn" tool which has a simpler command-line
syntax and can also run IgBLAST in parallel.

The results are parsed, postprocessed and printed as a tab-separated table
to standard output.

Postprocessing includes:

- The CDR3 is detected by using a regular expression
- The leader is detected within the sequence before the found V gene (by
  searching for the start codon).
- If the V sequence hit starts not at base 1 in the reference, it is extended
  to the left.
"""
import sys
import os
import shutil
import time
import multiprocessing
import subprocess
from contextlib import ExitStack
from io import StringIO
from itertools import islice
import hashlib
import errno
import pkg_resources
import logging
import tempfile
import json
import gzip

from xopen import xopen

from sqt import SequenceReader
from sqt.utils import available_cpu_count
from sqt.dna import nt_to_aa

from .utils import get_cpu_time, SerialPool
from .parse import TableWriter, IgBlastParser, ExtendedIgBlastParser
from .species import cdr3_start, cdr3_end


logger = logging.getLogger(__name__)


def add_arguments(parser):
	arg = parser.add_argument
	arg('--threads', '-t', '-j', type=int, default=available_cpu_count(),
		help='Number of threads. Default: no. of available CPUs (%(default)s)')
	arg('--penalty', type=int, choices=(-1, -2, -3, -4), default=None,
		help='BLAST mismatch penalty (default: -1)')
	arg('--species', default=None,
		help='Tell IgBLAST which species to use. Note that this setting does '
			'not seem to have any effect since we provide our own database to '
			'IgBLAST. Default: Use IgBLAST’s default')
	arg('--sequence-type', default='Ig', choices=('Ig', 'TCR'),
		help='Sequence type. Default: %(default)s')
	arg('--raw', metavar='FILE', help='Write raw IgBLAST output to FILE '
			'(add .gz to compress)')
	arg('--limit', type=int, metavar='N',
		help='Limit processing to first N records')
	arg('--rename', default=None, metavar='PREFIX',
		help='Rename reads to PREFIXseqN (where N is a number starting at 1)')
	arg('--stats', metavar='FILE',
		help='Write statistics in JSON format to FILE')

	arg('database', help='Database directory with V.fasta, D.fasta, J.fasta.')
	arg('fasta', help='File with original reads')


class IgBlastCache:
	"""Cache IgBLAST results in ~/.cache"""

	binary = 'igblastn'

	def __init__(self):
		version_string = subprocess.check_output([IgBlastCache.binary, '-version'])
		self._hasher = hashlib.md5(version_string)
		cache_home = os.environ.get('XDG_CACHE_HOME', os.path.expanduser('~/.cache'))
		self._cachedir = os.path.join(cache_home, 'igdiscover')
		logger.info('Caching IgBLAST results in %r', self._cachedir)
		self._lock = multiprocessing.Lock()

	def _path(self, digest):
		"""Return path to cache file given a digest"""
		return os.path.join(self._cachedir, digest[:2], digest) + '.txt.gz'

	def _load(self, digest):
		try:
			with gzip.open(self._path(digest), 'rt') as f:
				return f.read()
		except FileNotFoundError:
			return None

	def _store(self, digest, data):
		path = self._path(digest)
		dir = os.path.dirname(path)
		os.makedirs(dir, exist_ok=True)
		with gzip.open(path, 'wt') as f:
			f.write(data)

	def retrieve(self, variable_arguments, fixed_arguments, blastdb_dir, fasta_str) -> str:
		hasher = self._hasher.copy()
		hasher.update(' '.join(fixed_arguments).encode())
		for gene in 'V', 'D', 'J':
			with open(os.path.join(blastdb_dir, gene + '.fasta'), 'rb') as f:
				hasher.update(f.read())
		hasher.update(fasta_str.encode())
		digest = hasher.hexdigest()
		data = self._load(digest)
		if data is None:
			full_arguments = [IgBlastCache.binary] + variable_arguments + fixed_arguments
			data = subprocess.check_output(full_arguments, input=fasta_str,
				universal_newlines=True)
			with self._lock:  # TODO does this help?
				self._store(digest, data)

		return data


def run_igblast(sequences, blastdb_dir, species, sequence_type, penalty=None, use_cache=True) -> str:
	"""
	Run the igblastn command-line program.

	sequences -- list of Sequence objects
	blastdb_dir -- directory that contains BLAST databases. Files in that
	directory must be databases created by the makeblastdb program and have
	names V, D, and J.

	Return IgBLAST’s raw output as a string.
	"""
	if sequence_type not in ('Ig', 'TCR'):
		raise ValueError('sequence_type must be "Ig" or "TCR"')
	variable_arguments = []
	for gene in 'V', 'D', 'J':
		variable_arguments += ['-germline_db_{gene}'.format(gene=gene),
			os.path.join(blastdb_dir, '{gene}'.format(gene=gene))]
	# An empty .aux suppresses a warning from IgBLAST. /dev/null does not work.
	empty_aux_path = pkg_resources.resource_filename('igdiscover', 'empty.aux')
	variable_arguments += ['-auxiliary_data', empty_aux_path]
	arguments = []
	if penalty is not None:
		arguments += ['-penalty', str(penalty)]
	if species is not None:
		arguments += ['-organism', species]
	arguments += [
		'-ig_seqtype', sequence_type,
		'-num_threads', '1',
		'-domain_system', 'imgt',
		'-num_alignments_V', '1',
		'-num_alignments_D', '1',
		'-num_alignments_J', '1',
		'-outfmt', '7 sseqid qstart qseq sstart sseq pident slen evalue',
		'-query', '-',
		'-out', '-',  # write to stdout
	]
	fasta_str = ''.join(">{}\n{}\n".format(r.name, r.sequence) for r in sequences)

	if use_cache:
		global _igblastcache
		return _igblastcache.retrieve(variable_arguments, arguments, blastdb_dir, fasta_str)
	else:
		return subprocess.check_output(['igblastn'] + variable_arguments + arguments, input=fasta_str,
			universal_newlines=True)


def chunked(iterable, chunksize: int):
	"""
	Group the iterable into lists of length chunksize
	>>> list(chunked('ABCDEFG', 3))
	[['A', 'B', 'C'], ['D', 'E', 'F'], ['G']]
	"""
	chunk = []
	for it in iterable:
		if len(chunk) == chunksize:
			yield chunk
			chunk = []
		chunk.append(it)
	if chunk:
		yield chunk


class Runner:
	"""
	This is the target of a multiprocessing pool. The target needs to
	be pickleable, and because nested functions cannot be pickled,
	we need this separate class.

	It runs IgBLAST and parses the output for a list of sequences.
	"""
	def __init__(self, blastdb_dir, species, sequence_type, penalty, database):
		self.blastdb_dir = blastdb_dir
		self.species = species
		self.sequence_type = sequence_type
		self.penalty = penalty
		self.database = database

	def __call__(self, sequences):
		"""
		Return tuples (igblast_result, records) where igblast_result is the raw IgBLAST output
		and records is a list of ExtendedIgBlastRecord objects (the parsed output).
		"""
		igblast_result = run_igblast(sequences, self.blastdb_dir, self.species, self.sequence_type,
			self.penalty)
		sio = StringIO(igblast_result)
		if self.database is None:
			parser = IgBlastParser(sequences, sio)
		else:
			parser = ExtendedIgBlastParser(sequences, sio, self.database)
		records = list(parser)
		assert len(records) == len(sequences)
		return igblast_result, records


def makeblastdb(fasta, database_name):
	with SequenceReader(fasta) as fr:
		sequences = list(fr)
	if not sequences:
		raise ValueError("FASTA file {} is empty".format(fasta))
	process_output = subprocess.check_output(
		['makeblastdb', '-parse_seqids', '-dbtype', 'nucl', '-in', fasta, '-out', database_name],
		stderr=subprocess.STDOUT
	)
	shutil.copy(fasta, database_name + '.fasta')
	if b'Error: ' in process_output:
		raise subprocess.SubprocessError()


class Database:
	def __init__(self, path, sequence_type):
		"""path -- path to database directory with V.fasta, D.fasta, J.fasta"""
		self.path = path
		self.sequence_type = sequence_type
		with SequenceReader(os.path.join(path, 'V.fasta')) as sr:
			self._v_records = list(sr)
		self.v = self._records_to_dict(self._v_records)
		with SequenceReader(os.path.join(path, 'J.fasta')) as sr:
			self._j_records = list(sr)
		self.j = self._records_to_dict(self._j_records)
		self._cdr3_starts = dict()
		self._cdr3_ends = dict()
		for chain in ('heavy', 'kappa', 'lambda', 'alpha', 'beta', 'gamma', 'delta'):
			self._cdr3_starts[chain] = {name: cdr3_start(s, chain) for name, s in self.v.items()}
			self._cdr3_ends[chain] = {name: cdr3_end(s, chain) for name, s in self.j.items()}
		self.v_regions_nt, self.v_regions_aa = self._find_v_regions()

	@staticmethod
	def _records_to_dict(records):
		return {record.name: record.sequence.upper() for record in records}

	def v_cdr3_start(self, gene, chain):
		return self._cdr3_starts[chain][gene]

	def j_cdr3_end(self, gene, chain):
		return self._cdr3_ends[chain][gene]

	def _find_v_regions(self):
		"""
		Run IgBLAST on the V sequences to determine the nucleotide and amino-acid sequences of the
		FR1, CDR1, FR2, CDR2 and FR3 regions
		"""
		v_regions_nt = dict()
		v_regions_aa = dict()
		for record in igblast(self.path, self._v_records, self.sequence_type, threads=1):
			nt_regions = dict()
			aa_regions = dict()
			for region in ('FR1', 'CDR1', 'FR2', 'CDR2', 'FR3'):
				nt_seq = record.region_sequence(region)
				if nt_seq is None:
					break
				if len(nt_seq) % 3 != 0:
					logger.warning('Length %s of %s region in %r is not divisible by three; region '
						'info for this gene will not be available',
						len(nt_seq), region, record.query_name)
					# not codon-aligned, skip entire record
					break
				nt_regions[region] = nt_seq
				try:
					aa_seq = nt_to_aa(nt_seq)
				except ValueError as e:
					logger.warning('The %s region could not be converted to amino acids: %s',
						region, str(e))
					break
				if '*' in aa_seq:
					logger.warning('The %s region in %r contains a stop codon (%r); region info '
						'for this gene will not be available',
						region, record.query_name, aa_seq)
					break
				aa_regions[region] = aa_seq
			else:
				v_regions_nt[record.query_name] = nt_regions
				v_regions_aa[record.query_name] = aa_regions

		return v_regions_nt, v_regions_aa


def igblast(database, sequences, sequence_type, species=None, threads=None, penalty=None,
		raw_output=None):
	"""
	Run IgBLAST, parse results and yield (Extended-)IgBlastRecord objects.

	database -- Path to database directory with V./D./J.fasta files *or* a Database object.
	    If it is a path, then only IgBlastRecord objects are returned.
	sequences -- an iterable of Sequence objects
	sequence_type -- 'Ig' or 'TCR'
	threads -- number of threads. If None, then available_cpu_count() is used.
	raw_output -- If not None, raw IgBLAST output is written to this file
	"""
	if threads is None:
		threads = available_cpu_count()
	if isinstance(database, str):
		database_dir = database
		database = None
	else:
		database_dir = database.path
	with ExitStack() as stack:
		# Create the three BLAST databases in a temporary directory
		blastdb_dir = stack.enter_context(tempfile.TemporaryDirectory())
		for gene in ['V', 'D', 'J']:
			makeblastdb(os.path.join(database_dir, gene + '.fasta'), os.path.join(blastdb_dir, gene))

		chunks = chunked(sequences, chunksize=1000)
		runner = Runner(blastdb_dir, species=species, sequence_type=sequence_type, penalty=penalty, database=database)
		pool = stack.enter_context(multiprocessing.Pool(threads) if threads > 1 else SerialPool())
		for igblast_output, igblast_records in pool.imap(runner, chunks, chunksize=1):
			if raw_output:
				raw_output.write(igblast_output)
			yield from igblast_records


def main(args):
	global _igblastcache
	_igblastcache = IgBlastCache()
	database = Database(args.database, args.sequence_type)
	detected_cdr3s = 0
	writer = TableWriter(sys.stdout)
	start_time = time.time()
	with ExitStack() as stack:
		if args.raw:
			raw_output = stack.enter_context(xopen(args.raw, 'w'))
		else:
			raw_output = None
		sequences = stack.enter_context(SequenceReader(args.fasta))
		sequences = islice(sequences, 0, args.limit)

		n = 0  # number of records processed so far
		for record in igblast(database, sequences, sequence_type=args.sequence_type,
				species=args.species, threads=args.threads, penalty=args.penalty,
				raw_output=raw_output):
			n += 1
			if args.rename is not None:
				record.query_name = "{}seq{}".format(args.rename, n)
			d = record.asdict()
			if d['CDR3_aa']:
				detected_cdr3s += 1
			try:
				writer.write(d)
			except IOError as e:
				if e.errno == errno.EPIPE:
					sys.exit(1)
				raise
			if n % 100000 == 0:
				elapsed = time.time() - start_time
				logger.info('Processed {:10,d} sequences at {:.3f} ms/sequence'.format(n, elapsed / n * 1E3))
	elapsed = time.time() - start_time
	logger.info('Processed {:10,d} sequences at {:.1f} ms/sequence'.format(n, elapsed / n * 1E3))

	logger.info('%d IgBLAST assignments parsed and written', n)
	logger.info('CDR3s detected in %.1f%% of all sequences', detected_cdr3s / n * 100)
	if args.stats:
		stats = {'total': n, 'detected_cdr3s': detected_cdr3s}
		with open(args.stats, 'w') as f:
			json.dump(stats, f)
			print(file=f)



"""
Create and initialize a new analysis directory.
"""
import glob
import logging
import os
import os.path
import sys
import subprocess
import pkg_resources
from sqt import SequenceReader
from .config import Config

try:
	import tkinter as tk
	from tkinter import messagebox
	from tkinter import filedialog
except ImportError:
	tk = None

from xopen import xopen

logger = logging.getLogger(__name__)


def add_arguments(parser):
	parser.add_argument('--database', '--db', metavar='PATH', default=None,
		help='Directory with V.fasta, D.fasta and J.fasta files. If not given, a dialog is shown.')
	group = parser.add_mutually_exclusive_group()
	group.add_argument('--single-reads', default=None, metavar='READS',
		help='File with single-end reads (.fasta.gz or .fastq.gz)')
	group.add_argument('--reads1', default=None,
		help='First paired-end read file. The second is found automatically. '
			'Must be a .fastq.gz file. If not given, a dialog is shown.')
	parser.add_argument('directory', help='New analysis directory to create')


def launch(path):
	if hasattr(os, 'startfile'):
		os.startfile(path)
	elif sys.platform == 'linux':
		subprocess.call(['xdg-open', path])
	elif sys.platform == 'darwin':
		subprocess.call(['open', path])


class TkinterGui:
	"""Show a GUI for selecting reads and the database directory"""
	def __init__(self):
		if not tk:  # import failed
			raise ImportError()
		root = tk.Tk()
		root.withdraw()

	def yesno(self, title, question):
		return tk.messagebox.askyesno(title, question)

	def database_path(self, initialdir):
		path = tk.filedialog.askdirectory(
			title="Choose V/D/J database directory", mustexist=True,
			initialdir=initialdir)
		return path

	def reads1_path(self):
		path = tk.filedialog.askopenfilename(
			title="Choose first reads file",
			filetypes=[
				("Reads", "*.fastq.gz"),
				("Any file", "*")])
		return path

	def single_reads_path(self):
		path = tk.filedialog.askopenfilename(
			title="Choose single-end reads file",
			filetypes=[
				("Reads", "*.fasta *.fastq *.fastq.gz *.fasta.gz"),
				("Any file", "*")])
		return path

	def error(self, title, message):
		tk.messagebox.showerror(title, message)


"""
# Works, but let’s not introduce the PySide dependency for now.

def qt_path():
	import PySide
	from PySide.QtGui import QApplication
	from PySide.QtGui import QMessageBox, QFileDialog

	# Create the application object
	app = QApplication([])

	path = QFileDialog.getOpenFileName(None,
		"Open first reads file", '.', "FASTA/FASTQ reads (*.fastq *.fasta *.fastq.gz *.fasta.gz);; Any file (*)")
	# QMessageBox.information(None, 'Chosen file', path[0])
	return path[0]
"""


def is_1_2(s, t):
	"""
	Determine whether s and t are identical except for a single character of
	which one of them is '1' and the other is '2'.
	"""
	differences = 0
	one_two = {'1', '2'}
	for c1, c2 in zip(s, t):
		if c1 != c2:
			differences += 1
			if differences == 2:
				return False
			if {c1, c2} != one_two:
				return False
	return differences == 1


def guess_paired_path(path):
	"""
	Given the path to a file that contains the sequences for the first read in a
	pair, return the file that contains the sequences for the second read in a
	pair. Both files must have identical names, except that the first must have
	a '1' in its name, and the second must have a '2' at the same position.

	Return None if no second file was found or if there are too many candidates.

	>>> guess_paired_path('file.1.fastq.gz')  # doctest: +SKIP
	'file.2.fastq.gz'  # if that file exists
	"""
	base, name = os.path.split(path)
	glob_pattern = os.path.join(base, name.replace('1', '?'))
	paths = [p for p in glob.glob(glob_pattern) if is_1_2(p, path) and '_R1_' not in p]
	if len(paths) != 1:
		return None
	return paths[0]


class UnknownFileFormatError(Exception):
	pass


def file_type(path):
	"""
	Return 'fasta' or 'fastq' depending on file format. The file may optionally
	be gzip-compressed.
	"""
	if path.endswith('.gz'):
		file = xopen(path)
	else:
		file = open(path)
	with file as f:
		first_char = f.read(1)
		if first_char == '@':
			return 'fastq'
		elif first_char == '>':
			return 'fasta'
		else:
			raise UnknownFileFormatError('Cannot recognize format. File starts with neither ">" nor "@"')


def try_open(path):
	try:
		with open(path) as f:
			pass
	except OSError as e:
		logger.error('Could not open %r: %s', path, e)
		sys.exit(1)


def read_and_repair_fasta(path):
	"""
	Read a FASTA file and make sure it is suitable for use with makeblastdb.
	It repairs the following issues:

	- If a record is empty, it is skipped
	- If a record name occurs more than once, the second record name gets a suffix
	- If a sequence occurs more than once, occurrences after the first are skipped
	"""
	with SequenceReader(path) as sr:
		records = list(sr)

	names = set()
	sequences = dict()
	for r in records:
		r.sequence = r.sequence.upper()
		if len(r.sequence) == 0:
			logger.info("Record %r is empty, skipping it.", r.name)
			continue
		name = r.name
		i = 0
		while name in names:
			i += 1
			name = r.name + '_{}'.format(i)
		if name != r.name:
			logger.info('Record name %r replaced with %r because it occurs more than once', r.name, name)
		if r.sequence in sequences:
			logger.info('Skipping %r because it contains the same sequence as %r',
				r.name, sequences[r.sequence])
			continue
		sequences[r.sequence] = name
		names.add(name)
		r.name = name
		yield r


def main(args):
	if ' ' in args.directory:
		logger.error('The name of the analysis directory must not contain spaces')
		sys.exit(1)

	if os.path.exists(args.directory):
		logger.error('The target directory {!r} already exists.'.format(args.directory))
		sys.exit(1)

	# If reads files or database were not given, initialize the GUI
	if (args.reads1 is None and args.single_reads is None) or args.database is None:
		try:
			gui = TkinterGui()
		except ImportError:  # TODO tk.TclError cannot be caught when import of tk fails
			logger.error('GUI cannot be started. Please provide reads1 file '
				'and database directory on command line.')
			sys.exit(1)
	else:
		gui = None

	# Find out whether data is paired or single
	assert not (args.reads1 and args.single_reads)
	if args.reads1 is None and args.single_reads is None:
		paired = gui.yesno('Paired end or single-end reads',
			'Are your reads paired and need to be merged?\n\n'
			'If you answer "Yes", next select the FASTQ files '
			'with the first of your paired-end reads.\n'
			'If you answer "No", next select the FASTA or FASTQ '
			'file with single-end reads.')
		if paired is None:
			logger.error('Cancelled')
			sys.exit(2)
	else:
		paired = bool(args.reads1)

	# Assign reads1 and (if paired) also reads2
	if paired:
		if args.reads1 is not None:
			reads1 = args.reads1
			try_open(reads1)
		else:
			reads1 = gui.reads1_path()
			if not reads1:
				logger.error('Cancelled')
				sys.exit(2)
		reads2 = guess_paired_path(reads1)
		if reads2 is None:
			logger.error('Could not determine second file of paired-end reads')
			sys.exit(1)
	else:
		if args.single_reads is not None:
			reads1 = args.single_reads
			try_open(reads1)
		else:
			reads1 = gui.single_reads_path()
			if not reads1:
				logger.error('Cancelled')
				sys.exit(2)

	if args.database is not None:
		dbpath = args.database
	else:
		# TODO as soon as we distribute our own database files, we can use this:
		# database_path = pkg_resources.resource_filename('igdiscover', 'databases')
		databases_path = None
		dbpath = gui.database_path(databases_path)
		if not dbpath:
			logger.error('Cancelled')
			sys.exit(2)

	database = dict()
	for g in ['V', 'D', 'J']:
		path = os.path.join(dbpath, g + '.fasta')
		if not os.path.exists(path):
			logger.error(
				'The database directory %r must contain the three files '
				'V.fasta, D.fasta and J.fasta', dbpath)
			logger.error(
				'A dummy D.fasta is necessary even if analyzing light chains (see manual)')
			sys.exit(2)
		database[g] = list(read_and_repair_fasta(path))

	# Create the directory
	try:
		os.mkdir(args.directory)
	except OSError as e:
		logger.error(e)
		sys.exit(1)

	def create_symlink(readspath, dirname, target):
		gz = '.gz' if readspath.endswith('.gz') else ''
		if not os.path.isabs(readspath):
			src = os.path.relpath(readspath, dirname)
		else:
			src = readspath
		os.symlink(src, os.path.join(dirname, target + gz))

	if paired:
		create_symlink(reads1, args.directory, 'reads.1.fastq')
		create_symlink(reads2, args.directory, 'reads.2.fastq')
	else:
		try:
			target = 'reads.' + file_type(reads1)
		except UnknownFileFormatError:
			logger.error('Cannot determine whether reads file is FASTA or FASTQ')
			sys.exit(1)
		create_symlink(reads1, args.directory, target)

	# Write the configuration file
	configuration = pkg_resources.resource_string('igdiscover', Config.DEFAULT_PATH).decode()
	with open(os.path.join(args.directory, Config.DEFAULT_PATH), 'w') as f:
		f.write(configuration)

	# Create database files
	database_dir = os.path.join(args.directory, 'database')
	os.mkdir(database_dir)
	for gene in ['V', 'D', 'J']:
		with open(os.path.join(database_dir, gene + '.fasta'), 'w') as db_file:
			for record in database[gene]:
				print('>{}\n{}'.format(record.name, record.sequence), file=db_file)

	if gui is not None:
		# Only suggest to edit the config file if at least one GUI dialog has been shown
		if gui.yesno('Directory initialized',
				'Do you want to edit the configuration file now?'):
			launch(os.path.join(args.directory, Config.DEFAULT_PATH))
	logger.info('Directory %s initialized.', args.directory)
	logger.info('Edit %s/%s, then run "cd %s && igdiscover run" to start the analysis', args.directory, Config.DEFAULT_PATH, args.directory)



"""
Find V gene sister sequences shared by multiple libraries.
"""
import logging
from collections import Counter
import pandas as pd

logger = logging.getLogger(__name__)


def add_arguments(parser):
	arg = parser.add_argument
	arg('--minimum-frequency', '-n', type=int, metavar='N',
		default=None,
		help='Minimum number of datasets in which sequence must occur (default is no. of files divided by two)')
	arg('--minimum-db-diff', '-b', type=int, metavar='DIST', default=1,
		help='Use only sequences that have at least DIST differences to the database sequence. Default: %(default)s')
	arg('tables', metavar='DISCOVER.TAB',
		help='Table created by the "discover" command (give at least two)', nargs='+')


def main(args):
	if args.minimum_frequency is None:
		minimum_frequency = max((len(args.tables) + 1) // 2, 2)
	else:
		minimum_frequency = args.minimum_frequency
	logger.info('Minimum frequency set to %s', minimum_frequency)

	# Read in tables
	tables = []
	for path in args.tables:
		table = pd.read_csv(path, sep='\t')
		table = table[table.database_diff >= args.minimum_db_diff]
		table = table.dropna()
		tables.append(table)
		if len(table) == 0:
			logger.warn('Table read from %r is empty after filtering out sequences with database diff >= %s.', path, args.minimum_db_diff)

	# Count V sequence occurrences
	counter = Counter()
	for table in tables:
		counter.update(set(table.consensus))

	# Find most frequent occurrences and print result
	print('count', 'gene', 'database_diff', 'sequence', 'names', sep='\t')
	for sequence, frequency in counter.most_common():
		if frequency < minimum_frequency:
			break
		names = []
		gene = None
		for table in tables:
			matching_rows = table[table.consensus == sequence]
			if matching_rows.empty:
				continue
			names.extend(matching_rows.name)
			if gene is None:
				row = matching_rows.iloc[0]
				gene = row.gene
				database_diff = row.database_diff
				#shm = row['V_SHM']
		print(frequency, gene, database_diff, sequence, *names, sep='\t')



import random
from igdiscover.trie import Trie
from igdiscover.group import hamming_neighbors
from sqt.align import hamming_distance


def random_nt(length):
	return ''.join(random.choice('ACGT') for _ in range(length))


class TestTrie:
	LENGTHS = (0, 1, 2, 3, 4, 5, 6, 10, 12, 15, 20)

	def setup(self):
		self.strings = set()
		self.trie = Trie()
		# Create a set and a Trie both containing the same strings
		for _ in range(80):
			for length in self.LENGTHS:
				s = random_nt(length)
				self.strings.add(s)
				self.trie.add(s)

	def test_empty_string(self):
		trie = Trie()
		trie.add('')
		assert '' in trie
		assert 'A' not in trie

	def test_contains(self):
		for s in self.strings:
			assert s in self.trie

		for length in self.LENGTHS:
			for _ in range(min(100, 4**length)):
				s = random_nt(length)
				assert (s in self.strings) == (s in self.trie)

	def test_len(self):
		assert len(self.strings) == len(self.trie)

	def naive_has_similar(self, s, distance):
		for t in self.strings:
			if len(t) != len(s):
				continue
			if hamming_distance(t, s) <= distance:
				return True
		return False

	def naive_find_all_similar(self, s, distance):
		for t in self.strings:
			if len(t) != len(s):
				continue
			if hamming_distance(t, s) <= distance:
				yield t

	def test_has_similar(self):
		for s in self.strings:
			assert self.trie.has_similar(s, 0)
			assert self.trie.has_similar(s, 1)
			assert self.trie.has_similar(s, 2)

		for base in self.strings:
			for modified in hamming_neighbors(base):
				assert self.trie.has_similar(modified, 1)

		for errors in range(4):
			for length in self.LENGTHS:
				for _ in range(min(100, 4**length)):
					s = random_nt(length)
					assert self.naive_has_similar(s, errors) == self.trie.has_similar(s, errors)

	def test_find_all_similar(self):
		t = Trie()
		t.add('ACGT')
		result = list(t.find_all_similar('ACGG', 1))
		assert set(result) == frozenset(('ACGT',))

		for s in self.strings:
			result = list(self.trie.find_all_similar(s, 0))
			assert result == [s]

			for errors in range(1, 4):
				result = list(self.trie.find_all_similar(s, errors))
				assert s in result

		for base in self.strings:
			for modified in hamming_neighbors(base):
				assert base in self.trie.find_all_similar(modified, 1)

		for errors in range(1, 3):
			for length in self.LENGTHS:
				for _ in range(min(100, 4**length)):
					s = random_nt(length)
					expected = set(self.naive_find_all_similar(s, errors))
					assert expected == set(self.trie.find_all_similar(s, errors))


def main():
	import sys
	n = int(sys.argv[1])
	dist = int(sys.argv[2])

	strings = [random_nt(13) for _ in range(n)]
	print('created random sequences')

	if sys.argv[3] == 'x':
		def naive_has_similar(t):
			for s in strings:
				if hamming_distance(s, t) <= dist:
					return True
			return False

		hs = 0
		for s in strings:
			hs += int(naive_has_similar(s[1:] + s[0]))
		print('hs:', hs, 'out of', len(strings))

	else:
		trie = Trie()
		for s in strings:
			trie.add(s)
		print('created trie')
		hs = 0
		for s in strings:
			hs += int(trie.has_similar(s[1:] + s[0], dist))

		print('hs:', hs, 'out of', len(strings))


if __name__ == '__main__':
	main()
